//===------------- DefUseSites_Fer.cpp - Printing prototype ---------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it stores the definition and use sites of virtual registers.
//
// Date: June 12th, 2006
//
//===----------------------------------------------------------------------===//
//
// This class implements a mapping of virtual registers to instructions.
// For each virtual, this class gives its definition site, which is unique,
// due to the SSA property, and all the use sites. A definition site is the
// instruction where the register is defined, and a use site is an
// instruction where the register is used. Such information allows to
// compute a weighting function used to choose the best candidate for
// spilling. The weighting function is given by the following algorithm:
//
// Instruction def_i := definition_site(v);
//
// Unsigned w := 1 + 10 ** loop_info.nested_depth(def_i);
// 
// foreach ( Instruction i in use_sites(v) )
// |
// |    w := w + 1 + 10 ** loop_info.nested_depth(i);
// |
// end foreach
//
// v2w(v) := w;
//
//===----------------------------------------------------------------------===//

#ifndef DEF_USE_SITES_FER_H
#define DEF_USE_SITES_FER_H

#include "llvm/PassSupport.h"
#include "llvm/fernando/CriticalEdgeRemoval_Fer.h"

namespace llvm {

    class DefUseSites_Fer : public MachineFunctionPass {

        public:

        /// reporst that this pass does not change anything on the control
        /// flow graph.
        virtual void getAnalysisUsage(AnalysisUsage &AU) const {
            AU.addRequired<CriticalEdgeRemoval_Fer>();
            AU.addRequired<LoopInfo>();
            AU.setPreservesAll();
        }


        /// Informs the name of the pass, so it can appear in tools like
        /// llc.
        virtual const char * getPassName() const {
            return "Virtual to def/use mapping - Fernando.";
        }


        /// Returns a vector containing all the instructions where the given
        /// temporary is used.
        std::vector< MachineInstr * > &
        get_use_sites(const unsigned v_reg);


        /// This method returns true if the given operand has a definition
        /// site marked in this class; it returns false otherwise.
        const bool has_def_site(const unsigned v_reg) const;


        /// Returns the instruction where the temporary is defined. This
        /// instruction is unique, because the target program is expected to
        /// be in the SSA-form.
        const MachineInstr * get_def_site(const unsigned v_reg);


        /// Returns the number of the block where the machine operand was
        /// defined.
        int get_def_block(const unsigned v_reg);


        /// This is a light-weighted version of the get_weight method that
        /// reads operand objects. It will not check if the parameter is a
        /// valid register.
        const unsigned get_weight(const unsigned v_reg);


        /// Prints in the standard error device all the information contained
        /// in this data structure.
        virtual void dump() const;

        private:

        /// This maps basic block numbers to the nested depth of that basic
        /// block. For instance, if the machine basic block 2 has is in a
        /// loop nested three times, loop_depth[2] = 3;
        std::vector<unsigned> loop_depth;


        /// Mapping of registers to definition sites. Because in LLVM any
        /// register can be mapped to an unsigned integer, unsigned is
        /// used as the type for the keys.
        std::vector< const MachineInstr * > v2def;


        /// Mapping of registers to use sites. Contrary to definition, a
        /// virtual register can be used multiple times in the same program.
        std::vector< std::vector< MachineInstr * > > v2uses;


        /// Target architecture specific information can be accessed through
        /// this interface. It is used to print the name of instructions. If
        /// the dump method is removed, this variable can safely be
        /// removed.
        const TargetMachine * target_machine;


        /// This method visits a basic block and collects all the uses and defs
        /// present in this basic block.
        void visit_basic_block(MachineBasicBlock & mbb);


        /// Free the memory, so the operating system can use it in other
        /// applications.
        void releaseMemory();


        /// runOnMachineFunction - collects the use/def sites of each register,
        /// and computes their weights.
        bool runOnMachineFunction(MachineFunction & mf);


        /// Computes the e-th power of n, that is, n**e;
        const unsigned pow(unsigned n, unsigned e) const;


        /// Adds to the loop_depth vector the nested factor of the basic block
        /// mbb. The nested factor is computed by the LoopInfo analysis, which
        /// is available in the li variable.
        void set_loop_depth(MachineBasicBlock & mbb, LoopInfo & li);

    };  // end of EdgeLiveness_Fernando declaration

} // end llvm namespace

#endif
