//===------------- PhiCollector_Fer.cpp - Printing prototype --------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it groups virtual registers that are related by phi functions.
//
// Date: August 1th, 2006
//
//===----------------------------------------------------------------------===//
//
// This class groups together virtual registers that are related by phi
// functions in the control flow graph. Such virtuals describe the same
// variable in the original program. A good register allocator should try to
// assign the same physical register to these virtuals. For instance, if
// v1 := PHI (v2, v3, v3), then {v1, v2, v3} are phi related virtuals.
//
//===----------------------------------------------------------------------===//

#ifndef PHI_COLLECTOR_H
#define PHI_COLLECTOR_H

#include "llvm/PassSupport.h"

namespace llvm {

  class PhiCollector_Fer : public MachineFunctionPass {

    public:

    /// reporst that this pass does not change anything on the control flow
    /// graph.
    virtual void getAnalysisUsage(AnalysisUsage & AU) const {
        AU.setPreservesAll();
    }


    /// Informs the name of the pass, so it can appear in tools like llc.
    virtual const char * getPassName() const {
        return "Groups phi-related virtuals - Fernando.";
    }


    /// This method returns true if the given virtual (or, for the same effect,
    /// any other phi-related virtual) has been mapped to a stack slot.
    bool has_stack_slot(unsigned v);


    /// This method returns the stack slot that has been assigned to virtual
    /// v, if it has been spilled to memory. All the phi-related virtuals have
    /// the same stack slot.
    int get_stack_slot(unsigned v);


    /// This method maps the virtual v to the stack slot s. It also has the
    /// effect of assigning s to all other virtuals that are phi-related to v.
    /// If v has already been mapped to a stack slot, the function will produce
    /// an error.
    void map_to_stack_slot(unsigned v, int s);


    /// This method returns true if v_reg is part of a phi function somewhere
    /// in the control flow graph, and false otherwise. 
    bool is_phi_related(unsigned v_reg);


    /// This method adds "new_reg" to the set of related values of "v_reg".
    void add_to_related_set(unsigned v_reg, unsigned new_reg);


    /// Prints the set of phi-related virtuals
    void dump();


    /// This method reads an instruction, and adds all the virtuals in this
    /// instruction to the same related set. If the instruction is not a phi
    /// function, an error is returned.
    void add_as_related(const MachineInstr & phi);


    /// runOnMachineFunction - Groups together the phi related virtuals.
    bool runOnMachineFunction(MachineFunction & mf);

    private:

    /// For each cell (i, j) of this matrix, (i, j) will contain true if the
    /// virtuals i + first_virtual, and j + first_virtual are related by a
    /// phi function, or are transitivelly related. "first_virtual" is given by
    /// MRegisterInfo::FirstVirtualRegister.
    std::vector< std::vector<bool> > related;


    /// This vector maps virtuals to stack slots. All the phi-related virtuals
    /// are mapped to the same stack slot.
    std::vector<int> v2s;


    /// Free the memory, so the operating system can use it in other
    /// applications.
    void releaseMemory() {
        this->related.clear();
        this->v2s.clear();
    }

  };

}

#endif
