//===-------------------- PhiEq_Fer.cpp - PHI Matrix ----------------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This class represents the phi functions in the header of a basic block.
//
// Date: Jun 30th, 2006
//
//===----------------------------------------------------------------------===//
//
// This class implements a phi equation. A phi matrix equation such as
// A <=phi= M is an abstraction used to represent the phi functions that are
// present in the beginning of a basic block. The vector in the left side
// denotes the variables defined by the phi functions. The matrix in the right
// represents the parameters of the phi functions. For each predecessor basic
// block, the matrix M contains a column, that hold the variables alive if the
// flow reaches the current block coming from that predecessor. Ex.:
//
// a       | a_1 a_2 a_3 |
// b =phi= | b_1 b_2 b_3 |
// c       | c_1 c_2 c_3 |
//
// This matrix represents the three phi functions: a = (a1,a2,a3),
// b = (b1,b2,b3) and c = (c1,c2,c3).
//
//===----------------------------------------------------------------------===//

#ifndef PHI_EQUATION_H
#define PHI_EQUATION_H

#include <map>
#include <sstream>

#include "llvm/fernando/RegMapping_Fer.h"
#include "llvm/fernando/NeatPrint_Fer.h"

namespace llvm {

  class PhiEq_Fer {

  public:

    /// Constructor method. Creates a new matrix equation and fills it up with
    /// the phi functions defined in the given basic block. This method
    /// does not fills the virtual mapping in the phi equation.
    PhiEq_Fer(const MachineBasicBlock & mbb);
 

    /// Second constructor method. Creates a new matrix equation and fills it
    /// up with the phi functions defined in the given basic block. The
    /// difference for the first constructor is that, in this case, the values
    /// used to fill the phi function are given by the register mapping
    /// object.
    PhiEq_Fer(
        const MachineBasicBlock & mbb,
        const VirtRegMap & vrm,
        RegMapping_Fer & rm
    );
 

    /// Third constructor method. It creates a new phi equation from a second
    /// one. The valid vector marks which phi functions in ph must be added to
    /// the new equation.
    PhiEq_Fer(PhiEq_Fer & ph, std::vector<bool> valid);
 

    /// Each column of the phi matrix represents an execution path. This method
    /// returns the set of variables defined in the execution path from the
    /// pred block and the current block.
    std::vector<unsigned> & get_use_p_vec(const MachineBasicBlock & pred);


    /// This method swaps the two virtual registers in positions i1 and i2 of
    /// the vector that reaches this phi equation flowing from block pred.
    void swap_virtuals(
        unsigned i1,
        unsigned i2,
        const MachineBasicBlock & pred
    );


    /// This method returns the left side of the matrix equation, that is, the
    /// vector of variables defined by the set of phi functions that this
    /// equation represents.
    const std::vector<unsigned> & get_def_phys();


    /// Liberates memory to the operating system. This method must be
    /// invoked after this object is no longer necessary.
    ~PhiEq_Fer() {
        this->def_phys.clear();
        this->def_virt.clear();
        this->prm_phys.clear();
        this->prm_virt.clear();
        this->col2mbb.clear();
    }


    /// This method returns the number of phi functions present in this
    /// matrix equation. Each phi function defines a different variable.
    unsigned number_phi_functions() const;


    /// This method returns the number of predecessor blocks that gave origin
    /// to this matrix equation. Each predecessor block contains the
    /// live ranges of the variables represented in one of the columns of the
    /// matrix equation.
    unsigned ncols() const;


    /// Returns a string describing this phi matrix:
    /// For this set of phi functions: a = (a1,a2,a3), b = (b1,b2,b3) and
    /// c = (c1,c2,c3), the following matrix will be printed:
    ///
    /// a <= [ a_1 a_2 a_3 ]
    /// b <= [ b_1 b_2 b_3 ]
    /// c <= [ c_1 c_2 c_3 ]
    std::string to_string();


    /// This method prints a formated string. The string will be printed on a
    /// box of 'packSize'. The extra spaces will be filled with c.
    std::string print_pck_str(std::string str, int packSize, char c);


    /// This method returns a string describing this phi matrix. The difference
    /// for the parameterless to_string method is that in this case the name
    /// of the register is printed, instead of its code.
    std::string to_string(MachineFunction & mf);


    /// This method returns the virtual register that is mapped to the given
    /// index in the definition vector of the phi equation.
    unsigned get_def_virt(unsigned index) const;


    /// This method returns the virtual register that is mapped to the given
    /// index in the use vector that corresponds to the given basic block.
    unsigned get_use_virt(unsigned index, const MachineBasicBlock & mbb);

  private:

    /// Internal representation of the matrix equation:
    /// The vector that holds the variables defined in the equation:
    std::vector<unsigned> def_phys;

    /// This vector contains the virtual registers that are in the def vector
    /// of the phi equation.
    std::vector<unsigned> def_virt;

    /// The matrix containing the parameters of the phi functions. This data
    /// structure maps a block number to the vector that contains the variables
    /// alive in the path coming from that block.
    std::map< unsigned, std::vector<unsigned> > prm_phys;

    /// This matrix contains the virtual registers that are in the use matrix
    /// of the phi equation.
    std::map< unsigned, std::vector<unsigned> > prm_virt;

    /// This maps each column of the phi matrix to the number of the basic
    /// block that has produced it.
    std::vector<unsigned> col2mbb;

  };

}

#endif
