//===---------------- RegClique_Fer.h - Set of Registers ------------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This class implements a set of registers. It distinguises between virtual
// and machine registes.
//
// Date: May 27th, 2006
//
//===----------------------------------------------------------------------===//
//
// This class implements a set of registers. It is mostly used to represent
// cliques of registers. A clique of register is a set of register whose live
// ranges overlap at the same point in the control flow graph of the SSA-form
// program.
//===----------------------------------------------------------------------===//

#ifndef REG_CLIQUE_FER_H
#define REG_CLIQUE_FER_H

#include "llvm/fernando/EdgeLiveness_Fernando.h"

namespace llvm {

class RegClique_Fer {

    public:

    /// Constructor method. It will initialize this register set with the out
    /// set of the given machine block. The out set of a basic block is the
    /// set of variables alive at the end of the basic block. It can be
    /// obtained by the union of live sets of all the edges that leave the
    /// basic block.
    RegClique_Fer(
                   MachineFunction & mf,
                   const MachineBasicBlock & mbb,
                   const EdgeLiveness_Fernando * edge_liveness
                  );


    /// Constructor method. It will initialize this register set with the
    /// virtual registers alive in the edge that links mbb1 to mbb2. Notice
    /// that if mbb2 is not a successor of mbb1, no register will be in this
    /// register set.
    RegClique_Fer(
                   MachineFunction & mf,
                   const MachineBasicBlock & mbb1,
                   const MachineBasicBlock & mbb2,
                   const EdgeLiveness_Fernando * edge_liveness
                  );


    /// Constructor method. It produces a set of registers that is a copy of
    /// the set initially given as parameter.
    RegClique_Fer::RegClique_Fer(RegClique_Fer & rc);


    /// Destructor method. It returns to the operating system the memory that
    /// was used to keep the set of registers.
    ~RegClique_Fer() {
        this->virt_reg.clear();
        this->phys_reg.clear();
        this->allocatable_regs.clear();
    }


    /// Returns a new set of registers that is an intersection between this
    /// set and the given set.
    RegClique_Fer * intersection(RegClique_Fer & rc) const;


    /// This method returns a reference to the vector that represents the
    /// virtual registers stored in this set. This vector cannot be modified
    /// outside this class.
    const std::vector<bool> & get_virtual_vector() const;


    /// This method fills the given vector with integer numbers describing
    /// the registers in the set. Any previous contents of the vector will be
    /// erase. The first elements will represent physical registers, because,
    /// in LLVM, physical registers are denoted by integer numbers from 0 to K,
    /// where K is the number of physical registers.
    void to_vector(std::vector<unsigned> & regs) const;


    /// This method fills the given vector with all the physical registers
    /// present in this set.
    void p2vector(std::vector<unsigned> & regs) const;


    /// This method fills the given vector with all the virtual registers
    /// present in the set.
    void v2vector(std::vector<unsigned> & regs) const;


    /// This method returns true if the register set contains the register
    /// given as parameter. It will return false otherwise. The caller must
    /// make sure that the parameter describes a valid register number.
    bool contains_reg(unsigned reg_num) const;


    /// This method adds the register given as parameter to this register set.
    void add_reg(unsigned reg_num);


    /// This method removes the register given as parameter from this register
    /// set.
    bool remove_reg(unsigned reg_num);


    /// Returns a string in the format (r1, r2, ..., rm)+(t1, t2, ..., tn),
    /// where ri represents a machine register, and t1 represents a temporary
    /// (also called virtual) register.
    std::string to_string() const;


    /// This method adds each register uses in the instruction to this set of
    /// registers. Both, explicit and implicit uses are added.
    void add_uses(const MachineInstr & mi);


    /// This method removes each register defined in the instruction to this
    /// set of registers. Both, explicit and implicit definitions are removed.
    void remove_definitions(const MachineInstr & mi);


    /// This method grows the size of the set of virtual registers. Notice that
    /// it only GROWS the size, it never decrease it. Therefore, if the new
    /// size is smaller than the old one, the size will not change.
    void grow_virtual(unsigned new_size);


    /// Reporst the size of this set. The size of the set of registers is the
    /// sum of the number of virtual registers, and machine registers stored
    /// in it.
    unsigned size() const {
        return this->virt_size + this->phys_size;
    }


    /// Returns the number of physical registers in this set.
    unsigned num_p_regs() const {
        return this->virt_size + this->phys_size;
    }


    /// Returns the number of virtual registers in this set.
    unsigned num_v_regs() const {
        return this->virt_size + this->phys_size;
    }

    private:

    /// Number of physical registers in this set.
    unsigned phys_size;

    /// Number of virtual registers in this set.
    unsigned virt_size;

    /// Internal representation of the set: one vector for virtual
    /// registers, and another vector for physical registers.
    /// This is the vector for virtual registers.
    std::vector<bool> virt_reg;

    /// Internal representation of the set: one vector for virtual
    /// registers, and another vector for physical registers.
    /// This is the vector for physical registers.
    std::vector<bool> phys_reg;

    /// Target architecture specific information can be accessed through this
    /// interface.
    const TargetMachine * target_machine;

    /// Information about the bank of registers implemented by the actual
    /// machine can be accessed through this interface.
    const MRegisterInfo * m_reg_info;

    /// This vector keeps track of which registers can be used for register
    /// allocation purposes. Some registers cannot be used to hold the value
    /// of temporaries. For example, the ESP register in X86, or the fp
    /// register in MIPS. The cell corresponding to such registers in the
    /// vector below will be false. The other cells will be true;
    std::vector<bool> allocatable_regs;

}; // class definition

} // namespace

#endif
