//===-------- RegMapping_Fer.cpp - Bank of Machine Registers ---------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This class represents the bank of machine registers for register allocation
// purposes.
//
// Date: May 27th, 2006
//
//===----------------------------------------------------------------------===//
//
// This class represents the bank of machine registers for register allocation
// purposes. Essentially, it marks which physical registers are being used,
// and which ones are free to use. For each physical register in use, it keeps
// track of which virtual registers they use.
//
//===----------------------------------------------------------------------===//

#ifndef REG_MAPPING_FER_H
#define REG_MAPPING_FER_H

#include "llvm/fernando/RegClique_Fer.h"
#include "llvm/fernando/IG_Builder_Fer.h"

namespace llvm {
 
  class RegMapping_Fer {

    public:

    /// Constructor method. Initializes the data structures. The machine
    /// function parameter is only used to provide initializing information,
    /// such as the number of virtual and physical registers.
    RegMapping_Fer(MachineFunction & mf, IG_Builder_Fer & ig);


    /// Constructor method. It produces a mapping that is a copy of
    /// the set initially given as parameter.
    RegMapping_Fer(RegMapping_Fer & rm);


    /// Destructor method. Gives back to the operating system the memory that
    /// was in use during the object's execution.
    ~RegMapping_Fer();


    /// This method assigns an empty machine register to v_reg, if one is
    /// available. If the assignment is successfully performed, the function
    /// will return the true value. False will be returned if there is no
    /// register available. If v_reg is already allocated to a machine
    /// register, the method will return true, and nothing else is done.
    void allocate_color(unsigned reg, unsigned p_reg);


    /// This method does the same thing as the method allocate_color, but it can
    /// only be used with spilled registers.
    void set_color_spilled_register(unsigned reg, unsigned p_reg);


    /// This method clears the machine register that has been assigned to
    /// reg. It has the effect of spilling reg, and after it is called,
    /// any register from the same class as reg will be able to be allocated.
    /// That is, after liberate_color is called, allocate_color
    /// will return true if the parameters are from the same class of
    /// registers.
    void liberate_color(unsigned reg);


    /// This method returns the virtual register that was assigned to the given
    /// physical register. If the physical register does not contain any
    /// virtual register, the method will return egMapping_Fer::NO_REG. It is
    /// possible that the given physical register is a pre-colored register.
    /// In this case, the physical value itself will be returned.
    unsigned get_assigned_data(unsigned p_reg);


    /// This method returns the physical register that was allocated to the
    /// virtual register. If the virtual register has not been allocated to
    /// any register, it will return egMapping_Fer::NO_REG.
    unsigned get_physical_location(unsigned v_reg);


    /// This method cleans the mapping between physical registers and
    /// virtual registers, and if the registers in the given clique have
    /// already been allocated, new values will be assigned to the bank of
    /// machine registers to reflect the allocation. This method is generally
    /// called before starting the allocation of a basic block. The sp value
    /// is used to distinguish spilled virtuals from non-spilled virtuals. It
    /// avoids that spilled virtuals in the first clique of the vector be added
    /// as alive in the input set of the basic block.
    void reset_physical_mapping(unsigned sp, RegClique_Fer & ql);


    /// This method returns a string in the format "t1(r1), t2(r2), ... tm(rm),
    /// where virtual registers ti are allocated in machine registers ri.
    std::string virtual_to_string();


    /// This method returns a string in the format "r1(t1), r2(t2), ... rm(tm),
    /// where physical registers ri is holding the value of virtual register
    /// ti.
    std::string physical_to_string();


    /// This method returns a register that can be assigned to the given
    /// virtual register. It will return RegMapping_Fer::NO_REG if there
    /// is no register available. Notice that is is possible that there will
    /// be a register available, and even though this method returns NO_REG.
    /// This will happen if all the available registers are pre-colored
    /// registers that interfere with this one is some point of the control
    /// flow graph.
    unsigned get_free_register(unsigned v_reg);


    /// This method is very similar to get_free_register. Thid different is
    /// that here even a pre-colored register that interferes with v_reg can
    /// be returned.
    unsigned get_unused_reg(unsigned v_reg);


    /// Increase the number of virtual registers stored in the object by one.
    /// This method returns the new number of virtual registers.
    unsigned grow();


    /// This constant is used to identify which physical registers are free,
    /// and can be assigned to virtual registers.
    static const unsigned NO_REG = 65535;


    /// This value represents spilled virtuals.
    static const unsigned SP_REG = 65534;


    /// Informs the last register not yet spilled.
    unsigned get_last_non_spill() {
        return  this->last_register_non_spill;
    }


    /// This method verifies if p_reg can be used to allocated v_reg. This will
    /// be the case if both registers belong to the same class.
    bool compatible_class(unsigned v_reg, unsigned p_reg);

    private:

    /// This number denotes the last register that is not an spilled register.
    /// Any register whose number is greater than last_register_non_spill will
    /// be a virtual register created to break the live range of a spilled
    /// register.
    unsigned last_register_non_spill;


    /// This variable holds the interferences between physical and virtual
    /// registers in all the basic blocks. It avoids assigning a color to a
    /// virtual and later having to spill this virtual because this color
    /// is used by an overlapping physical (pre-colored) register.
    IG_Builder_Fer * ig;


    /// This is the table that maps physical registers to virtual registers.
    /// If p2v[p_reg] = NO_REG, than p_reg is not assigned to any
    /// virtual register.
    std::vector<unsigned> p2v;


    /// This vector maps virtual registers to physical registers. A
    /// BankOfRegisters_Fer::NO_REG value indicates that the virtual register
    /// hasn't been assigned a value yet.
    std::vector<unsigned> v2p;


    /// This comment is exactly the one in llvm/CodeGen/MachineFunction.h:
    /// This is a new[]'d array of bools that is computed and set by the
    /// register allocator, and must be kept up to date by passes that run
    /// after register allocation (though most don't modify this).  This is
    /// used so that the code generator knows which callee save registers to
    /// save and for other target specific uses.
    bool *p_regs_ever_used;


    /// The bank of register keeps a reference to the machine function in
    /// order to get the architecture specific information.
    MachineFunction * machine_function;


    /// This method determines if p_reg is not currently allocated to any
    /// virtual register. In order to verify this, it does not suffice to
    /// check if the cell of p_reg in the physical register table contains the
    /// NO_REG value, because physical registers contains alias. For instance,
    /// in the X86 architecture, the physical registers AL, AX and EAX are all
    /// alias, and if one of them is in use, the others are not free. If p_reg
    /// is not been used to allocate any virtual, the method returns true. It
    /// returns false otherwise.
    bool is_register_available(unsigned p_reg);

  };

} // end of llvm namespace

#endif
