//===----------- CfgV_Fernando.cpp - Printing prototype -------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it performs register allocation via choloring of chordal
// graphs.
//
// Date: June 7th, 2006
//
//===----------------------------------------------------------------------===//
//
// This file contains the implementation of a register allocation algorithm.
// Contrary to traditional RA approaches, this algorithm performs register
// assignment while the program is in SSA form, and uses parallel copies to
// eliminate phi functions, after a mapping between temporary and machine
// registers has been found. This algorithm is non-iterative, that is, it finds
// an optimal coloring of the interference graph in one scan of the control
// flow.
//
//===----------------------------------------------------------------------===//


#include <vector>
#include <sstream>                               // ostringstream (RegClique)
#include <iostream>                              // cerr

#include "VirtRegMap.h"
#include "llvm/CodeGen/Passes.h"                 // RegAllocChord_Fer
#include "llvm/Instructions.h"                   // MachineOperand, etc
#include "llvm/Analysis/LoopInfo.h"              // DefUseSites depends on it.
#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/CodeGen/MachineFunctionPass.h"    // MachineBasicBlock
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/ADT/DepthFirstIterator.h"         // df_ext_iterator
#include "llvm/ADT/Statistic.h"

#include "llvm/fernando/EdgeLiveness_Fernando.h"
#include "llvm/fernando/DefUseSites_Fer.h"
#include "llvm/fernando/IG_Builder_Fer.h"
#include "llvm/fernando/RegMapping_Fer.h"
#include "llvm/fernando/NeatPrint_Fer.h"
#include "llvm/fernando/PhiDeconstruction_Fer.h"
#include "llvm/fernando/KillingSites_Fer.h"

using namespace llvm;

namespace {

    Statistic<> numJoins("chordal", "Number of interval joins performed");

    Statistic<> numFolded
                ("chordal", "Number of loads/stores folded into instructions");

    class RegAllocChordal_Fer : public MachineFunctionPass {

    public:

    virtual void getAnalysisUsage(AnalysisUsage & AU) const {
        AU.addRequired<DefUseSites_Fer>();
        AU.addRequired<EdgeLiveness_Fernando>();
        AU.addRequired<IG_Builder_Fer>();
        MachineFunctionPass::getAnalysisUsage(AU);
    }

    /// The label of this pass is "Register allocation via coloring of
    /// chordal graphs."
    virtual const char * getPassName() const {
        return "Register allocation via coloring of chordal graphs.";
    }

    private:

    /// Maps virtual registers to physical registers and virtual registers
    ///  to stack slots. It is created and updated by a register allocator and
    /// then used by a machine code rewriter that adds spill code and rewrites
    /// virtual into physical register references. I am using this mapping
    /// instead of directly invoking the setReg() method because the spiller
    /// implements some optimizations on the final code.
    std::auto_ptr<VirtRegMap> vrm;


    /// The spiller generates load and store instructions due to spilled
    /// registers. It tries to reuse loads as much as possible, in order to
    /// avoid unnecessary instructions.
    std::auto_ptr<Spiller> spiller;


    /// This interface represents the machine function that is beeing
    /// compiled. Different data structures can be accessed thrugh this
    /// interface, e.g:
    /// const TargetMachine & tm = this->machine_function.getTarget();
    MachineFunction * machine_function;


    /// This variable contains liveness analysis information.
    EdgeLiveness_Fernando * edge_liveness;


    /// This variable maps virtual registers to the instructions where it
    /// is defined and used.
    DefUseSites_Fer * def_use_sites;


    /// This class is a interference graph. This interference graph only holds
    /// interferences between physical registers (the pre-colored registers)
    /// and virtual registers whose live ranges overlap.
    IG_Builder_Fer * ig;


    /// This data type keeps track of the mapping between virtual and
    /// physical registers.
    RegMapping_Fer * reg_mapping;


    /// This set contains the basic blocks that have already been visited
    /// by the register allocator.
    std::vector<bool> colored;


    /// This method erases any trace of the spilled register from the data
    /// structures. First, it removes the live range of the spilled virtual
    /// from the edge liveness results. Next, it removes the virtual from the
    /// list of cliques.
    void clean_live_range(
        KillingSites_Fer & ks,
        unsigned v_reg,
        const MachineBasicBlock & mbb
    );


    /// This method creates a new virtual register with the same class as
    /// reg_model;
    unsigned create_new_virtual_register(unsigned reg_model);


    /// After a virtual register is last used, the machine register assigned
    /// to it must be liberated, so it can be used to allocate other virtual
    /// registers. This method does exaclty this.
    void liberate_dead_uses(MachineInstr * mi, KillingSites_Fer & ks);


    /// It is possible that the code being compiled contains dead definitions,
    /// that is, a definition of a variable that is never used after been
    /// defined. In this case, the machine register used to hold the value
    /// of such definition can be liberated to allocate other live ranges.
    void liberate_dead_defs(MachineInstr * mi, KillingSites_Fer & ks);


    /// This method assigns a valid physical register to the virtual register.
    /// If there is not physical register available, it will spill one of the
    /// virtual registers, to make room for the virtual. The parameter position
    /// marks the index of the clique of live registers immediately after the
    /// instruction is executed.
    void allocate_defs(MachineInstr & mi, KillingSites_Fer & ks);


    /// This method finds physical registers for the virtual registers that
    /// have been spilled. Even though the virtual is saved in memory, it is
    /// necessary to find a register for it when it is used in an instruction.
    MachineInstr * allocate_spilled_uses
                              (MachineInstr * mi, KillingSites_Fer & ks);


    /// this method determines chooses a register to be spilled. The chosen
    /// register must be in the same register class as v_reg. The
    /// heuristics used to determine which register to spill is implemented
    /// in the DefUse class.
    unsigned choose_reg_to_spill(unsigned v_reg);


    /// This is the dual of the initialize_data_structures method. It will
    /// free all the memory used during register allocation.
    void delete_data_structures();


    /// This method initializes the data structures used during register
    /// allocation. 
    void initialize_data_structures(MachineFunction &mf);


    /// runOnMachineFunction - Invokes the register allocation algorithm.
    bool runOnMachineFunction(MachineFunction & machine_function);


    /// This method replace phi functions by parallel copies. Parallel
    /// copies are implemented with a mix of move and xor instructions
    void eliminate_phi_functions( MachineFunction & mf );


    /// This method performs register spilling. The parameter indicates a
    /// class of registers. The spilling algorithm must evict a register from
    /// the same class as the parameter. I am using the VirtRegMap class to
    /// insert loads and stores.
    void spill(
        unsigned sp_v,
        KillingSites_Fer & ks,
        const MachineBasicBlock & mbb
    );


    /// This method finds a mapping between virtual and machine registers
    /// in the basic block. The colored set points which nodes have
    /// already been visited in the depth first traversal of the
    /// dominance tree.
    void perform_local_allocation(MachineBasicBlock & mbb);


    /// This method converts the mapping from virtual register to physical
    /// register found by the perform_local_allocation method to a real
    /// assignment of physical registers in each operand of each instruction
    /// of the given basic block. Actually, it would be possible to implement
    /// this method directly in the perform_local_allocation function, but the
    /// separate implementation makes the program more readable.
    void hard_code_register_mapping(MachineFunction &mf);


    /// This method returns true if the given basic block is contained
    /// in the given set, and false otherwise.
    bool has_already_been_visited(MachineBasicBlock & mbb);

    };

}

//===--------------------------------------------------------------------------
// This method returns true if the given basic block is contained in the given
// set, and false otherwise.
//===--------------------------------------------------------------------------
bool RegAllocChordal_Fer::has_already_been_visited(MachineBasicBlock & mbb) {
    return this->colored[mbb.getNumber()];
}


//===--------------------------------------------------------------------------
// Traditional algorithms used to destroy phi functions normally replace them
// by copy instructions. Although the classical algorithms preserve the
// semantics of the SSA-form program, they may increase the number of registers
// found during the register assignment phase. To avoid this problem, we
// replace phi functions by parallel copies. Parallel copies are normally
// implemented with xor instructions (Remember the trick of swaping two
// variables without an auxiliary using xor?). The register swap is a costly
// operation, because it demands three instructions to be implemented;
// therefore, it should be avoided whenever possible. One way of avoiding
// swaps is by means of move instructions. Basically, swap operations are not
// necessary in every time that phi instructions are destroyed. This happens
// when the definition vector contains machine registers not used in the
// parameter vector. The algorithm shown below tries to use such optimization
// before inserting swap operations. It performs phi elimination in time
// linear in the size of the definition vector.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::eliminate_phi_functions( MachineFunction & mf ) {
    PhiDeconstruction_Fer phi_destructor
                     (* this->edge_liveness, * this->reg_mapping, * this->vrm);
    phi_destructor.runOnMachineFunction(mf);
}


//===--------------------------------------------------------------------------
// This method initializes the data structures used during register allocation.
// Among the variables initilized, are edge_liveness, and def_use_sites.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::initialize_data_structures(MachineFunction & mf) {
    // gets the liveness analysis information
    this->edge_liveness = & getAnalysis<EdgeLiveness_Fernando>();

    // gets the list of def/use sites
    this->def_use_sites = & getAnalysis<DefUseSites_Fer>();

    // builds the interference graph. Notice that this interference graph
    // only holds interferences between physical and virtual registers.
    this->ig = & getAnalysis<IG_Builder_Fer>();

    // gets information about target architecture
    this->machine_function =  & mf;

    // initializes the vector that marks visited nodes.
    MachineBasicBlock * last_block = mf.getLastBlock();
    unsigned num_blocks = last_block->getNumber() + 1;
    this->colored.assign(num_blocks, false);

    // initializes the mapping between virtual and physical registers. The
    // array mf::UsedPhysRegs will be set in the constructor of RegMapping.
    this->reg_mapping = new RegMapping_Fer(mf, * this->ig);

    // prepared the VirtRgMap to receive new data:
    this->vrm.reset(new VirtRegMap(mf));
    this->spiller.reset(createSpiller());
}


//===--------------------------------------------------------------------------
// This is the dual of the initialize_data_structures method. It will free
// all the memory used during register allocation.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::delete_data_structures() {
    // initializes the mapping between virtual and physical registers.
    // But don't delete the pass variables. This is the job of the pass
    // manager. If you delete them, the program will crash, and it will be
    // very, very hard to find the bug.
    delete this->reg_mapping;
    this->colored.clear();
}


//===--------------------------------------------------------------------------
// This method creates a new virtual register with the same class as reg_model.
//===--------------------------------------------------------------------------
unsigned RegAllocChordal_Fer::create_new_virtual_register(unsigned reg_model) {
    // Take the register class.
    const TargetRegisterClass *trc =
                this->machine_function->getSSARegMap()->getRegClass(reg_model);
    SSARegMap * srm = this->machine_function->getSSARegMap();
    return srm->createVirtualRegister(trc);
}


//===--------------------------------------------------------------------------
// This method determines a candidate to be sent to memory.
//===--------------------------------------------------------------------------
unsigned RegAllocChordal_Fer::choose_reg_to_spill(unsigned v_reg) {
    std::vector<unsigned> usefull_regs;

    // Take the register class.
    const TargetRegisterClass *trc =
                    this->machine_function->getSSARegMap()->getRegClass(v_reg);
    TargetRegisterClass::iterator trci =
                         trc->allocation_order_begin(* this->machine_function);
    TargetRegisterClass::iterator trce =
                           trc->allocation_order_end(* this->machine_function);

    for (; trci != trce; ++trci) {
        unsigned v_reg = this->reg_mapping->get_assigned_data(*trci);
        usefull_regs.push_back(v_reg);
    }

    unsigned best_cdd = RegMapping_Fer::NO_REG;
    unsigned best_w = 65535;
    for(unsigned u = 0; u < usefull_regs.size(); u++) {
        unsigned v_reg = usefull_regs[u];
        if(! this->vrm->hasStackSlot(v_reg)) {
            unsigned w = this->def_use_sites->get_weight(v_reg);
            if(w < best_w) {
                best_cdd = v_reg;
                best_w = w;
            }
        }
    }

    return best_cdd;
}


//===--------------------------------------------------------------------------
// This method performs register spilling. The parameter indicates a class of
// registers. The spilling algorithm must evict a register from the same class
// as the parameter. I am using the VirtRegMap class to place loads and stores.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::spill(
    unsigned v_reg,
    KillingSites_Fer & ks,
    const MachineBasicBlock & mbb
) {
    // First, find a place to store this register. Of course, this will be done
    // by the implementation of vrm. We just have to ask it.
    int slot = vrm->assignVirt2StackSlot(v_reg);
    unsigned p_reg = this->reg_mapping->get_physical_location(v_reg);   

    // Now, it is necessary to break the live range of the spilled register.
    // This is done by creating new virtual registers, and substituting the
    // spilled register by the new registers.
    MachineInstr * last_seen;
    std::vector< MachineInstr * > & use_sites =
                                     this->def_use_sites->get_use_sites(v_reg);
    for(unsigned u = 0; u < use_sites.size(); u++) {
        MachineInstr * mi = use_sites[u];
        if(mi == last_seen) {
            continue; // this happens when the same virtual is used multiple
                      // times in the same instruction.
        }
        unsigned new_reg = create_new_virtual_register(v_reg);
        if(mi->getParent()->getNumber() == ks.get_basic_block()->getNumber()) {
            ks.replace_used_reg(mi, new_reg, v_reg);
        }
        this->vrm->grow();
        this->reg_mapping->grow();
        this->vrm->assignVirt2StackSlot(new_reg, slot);
        for(unsigned t = 0; t < mi->getNumOperands(); t++) {
            MachineOperand & mo_aux = mi->getOperand(t);
            if(mo_aux.isRegister() && mo_aux.getReg() && mo_aux.isUse()) {
                if(mo_aux.getReg() == v_reg) {
                    mo_aux.setReg(new_reg);
                    this->reg_mapping->set_color_spilled_register
                                                              (new_reg, p_reg);
                }
            }
        }
        last_seen = mi;
    }

    // this method will clean the machine register, e.g. p_reg, that is been
    // currently used to hold the color of v_reg.
    this->reg_mapping->liberate_color(p_reg);

    clean_live_range(ks, v_reg, mbb);
}


//===--------------------------------------------------------------------------
// This method finds physical registers for the virtual registers that
// have been spilled. Even though the virtual is saved in memory, it is
// necessary to find a register for it when it is used in an instruction.
// The function only checks instructions that are not phi functions. If a phi
// function has a spilled use, this use will be allocated to an undefined
// register.
//===--------------------------------------------------------------------------
MachineInstr * RegAllocChordal_Fer::allocate_spilled_uses
                                   (MachineInstr * mi, KillingSites_Fer & ks) {
    if(mi->getOpcode() != TargetInstrInfo::PHI) {
        /// This set helps to avoid allocating the same use twice. Even if a
        /// virtual is used twice or more in an instruction, it only needs one
        /// physical register.
        std::set<unsigned> seen_regs;
        for(unsigned u = 0; u < mi->getNumOperands(); u++) {
            const MachineOperand & mo = mi->getOperand(u);
            if(mo.isUse() && mo.isRegister() && mo.getReg()) {
                if(MRegisterInfo::isVirtualRegister(mo.getReg())) {
                    unsigned v_reg = mo.getReg();
                    if(this->vrm->hasStackSlot(v_reg)) {
                        int slot = this->vrm->getStackSlot(v_reg);
                        // First, try to fold the memory reference into the
                        // instruction. If we can do this, we don't need to
                        // insert spill code.
                        const TargetMachine & target_machine =
                                           this->machine_function->getTarget();
                        const MRegisterInfo *ri =
                                              target_machine.getRegisterInfo();
                        MachineInstr * fmi =
                                            ri->foldMemoryOperand(mi, u, slot);
                        if(fmi) {
                            numFolded++;
                            MachineBasicBlock * mbb = mi->getParent();
                            //this->vrm->virtFolded(v_reg, mi, u, fmi);
                            ks.fold_inst(mi, fmi);
                            mi = mbb->insert(mbb->erase(mi), fmi);
                            u = 0;
                            seen_regs.clear();
                            continue;
                        }
                        // It was not possible to fold the spill into the
                        // instruction: a register must be fond.          
                        if(seen_regs.find(v_reg) != seen_regs.end()) {
                            continue;
                        } else {
                            seen_regs.insert(v_reg);
                        }
                        unsigned p_reg = reg_mapping->get_free_register(v_reg);
                        if(p_reg != RegMapping_Fer::NO_REG) {
                            this->reg_mapping->allocate_color(v_reg, p_reg);
                        } else {
                            unsigned p_reg =reg_mapping->get_unused_reg(v_reg);
                            if(p_reg != RegMapping_Fer::NO_REG) {
                                reg_mapping->allocate_color(v_reg, p_reg);
                            } else {
                                 unsigned spill_v = choose_reg_to_spill(v_reg);
                                unsigned p_reg =
                                   reg_mapping->get_physical_location(spill_v);
                                spill(spill_v, ks, * mi->getParent());
                                reg_mapping->allocate_color(v_reg, p_reg);
                            }
                        }
                    } 
                }
            }
        }
    }
    return mi;
}


//===--------------------------------------------------------------------------
// This method assigns a valid physical register to the virtual register. If
// there is not physical register available, it will spill one of the
// virtual registers, to make room for the virtual.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::allocate_defs(
    MachineInstr & mi,
    KillingSites_Fer & ks
) {
    // Gets the information about the given instruction:
    const TargetMachine & target_machine = this->machine_function->getTarget();
    const TargetInstrInfo & tii = * (target_machine.getInstrInfo());
    const TargetInstrDescriptor &tid = tii.get(mi.getOpcode());

    // If the instruction is a move, try to coalesce it:
    unsigned use_reg, def_reg;
    if(tii.isMoveInstr(mi, use_reg, def_reg)) {
        if(MRegisterInfo::isVirtualRegister(def_reg)) {
            unsigned p_reg;
            if(MRegisterInfo::isPhysicalRegister(use_reg)) {
                p_reg = use_reg;
            } else {
                p_reg = this->reg_mapping->get_physical_location(use_reg);
            }
            bool use_is_dead = ks.is_use_killing_site(& mi, use_reg);
            bool is_comp = this->reg_mapping->compatible_class(def_reg, p_reg);
            bool no_overlap = ! this->ig->is_connected(p_reg, def_reg);
            if( use_is_dead && is_comp && no_overlap ) {
                this->reg_mapping->allocate_color(def_reg, p_reg);
                numJoins++;
                return;
            }
        }
    }

    // First, handle the explicit definitions:
    for (unsigned i = 0; i != mi.getNumOperands(); i++) {
        const MachineOperand & mo = mi.getOperand(i);
        if(mo.isDef() && mo.isRegister() && mo.getReg()) {
            unsigned reg = mo.getReg();
            if(MRegisterInfo::isVirtualRegister(reg)) {
                unsigned p_reg = this->reg_mapping->get_free_register(reg);
                if(p_reg != RegMapping_Fer::NO_REG) {
                    this->reg_mapping->allocate_color(reg, p_reg);
                } else {
                    unsigned p_reg = this->reg_mapping->get_unused_reg(reg);
                    if(p_reg != RegMapping_Fer::NO_REG) {
                        this->reg_mapping->allocate_color(reg, p_reg);
                        spill(reg, ks, * mi.getParent());
                    } else {
                        unsigned spill_v = choose_reg_to_spill(reg);
                        unsigned p_reg =
                             this->reg_mapping->get_physical_location(spill_v);
                        spill(spill_v, ks, * mi.getParent());
                        this->reg_mapping->allocate_color(reg, p_reg);
                        if(this->ig->is_connected(p_reg, reg)) {
                            spill(reg, ks, * mi.getParent());
                        }
                    }
                }
            }
        } else {
            break;
        }
    }
    // Secondly, handle the implicit definitions:
    for (const unsigned *implicit_def = tid.ImplicitDefs;
                                               *implicit_def; ++implicit_def) {
        unsigned p_reg = *implicit_def;
        this->reg_mapping->allocate_color(p_reg, p_reg);
    }
}


//===--------------------------------------------------------------------------
// It is possible that the code being compiled contains dead definitions, that
// is, a definition of a variable that is never used after been defined. In
// this case, the machine register used to hold the value of such definition
// can be liberated to allocate other live ranges. This is what this method
// does.  The clique that it receives contains all the virtual and physical
// registers alive after the instructions is executed.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::liberate_dead_defs
                                   (MachineInstr * mi, KillingSites_Fer & ks) {
    ks.point_to_first_dead_def(mi);
    while(ks.has_dead_def()) {
        unsigned dead_def = ks.get_next_dead_def();
        this->reg_mapping->liberate_color(dead_def);
    }
}


//===--------------------------------------------------------------------------
// After a virtual register is last used, the machine register assigned to it
// must be liberated, so it can be used to allocate other virtual registers.
// This method does exaclty this. The clique that it receives contains all the
// virtual and physical registers alive after the instructions is executed.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::liberate_dead_uses
                                   (MachineInstr * mi, KillingSites_Fer & ks) {
    // liberate explicit uses that are not spills:
    ks.point_to_first_dead_use(mi);
    while(ks.has_dead_use()) {
        unsigned dead_use = ks.get_next_dead_use();
        this->reg_mapping->liberate_color(dead_use);
    }

    // liberate virtuals that contain spilled values:
    for (unsigned i = 0; i != mi->getNumOperands(); i++) {
        const MachineOperand & mo = mi->getOperand(i);
        if(mo.isRegister() && mo.isUse() && mo.getReg()) {
            if(MRegisterInfo::isVirtualRegister(mo.getReg())) {
                if(this->vrm->hasStackSlot(mo.getReg())) {
                    this->reg_mapping->liberate_color(mo.getReg());
                }
            }
        }
    }
}


//===--------------------------------------------------------------------------
// This method finds a mapping between virtual and machine registers in the
// basic block.
// The algorithm shown below finds a mapping between virtual registers and
// machine registers in linear time on the size of the interference graph.
// Notice that the interference graph is not really built. Also, notice that,
// due to properties of the SSA transformation, the algorithm is optimal if
// spills do not happen. In order to find an optimal assignment, the dominance
// tree is traversed using pre-order.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::perform_local_allocation(MachineBasicBlock & mbb) {
    // mark the node as already visited. Parts of the algorithm depends on
    // knowing if the node has been already visited.
    this->colored[mbb.getNumber()] = true;

    RegClique_Fer ql(*this->machine_function, mbb, this->edge_liveness);

    // This constructor modifies ql to contain the set of registers alive in
    // the beginning of the basic block.
    KillingSites_Fer ks(& mbb, ql);

    this->reg_mapping->reset_physical_mapping
                                 (this->reg_mapping->get_last_non_spill(), ql);

    // Allocates a physical register to each virtual register
    MachineBasicBlock::iterator mbbi = mbb.begin();
    while (mbbi != mbb.end()) {
        MachineInstr * mi = mbbi++;
        mi = allocate_spilled_uses(mi, ks);
        liberate_dead_uses(mi, ks);
        allocate_defs(*mi, ks);
        liberate_dead_defs(mi, ks);
        // Remember, liberate_dead_defs can delete the instruction, so, don't
        // use mi after this point.
    }
}


//===--------------------------------------------------------------------------
// This method erases any trace of the spilled register from the data
// structures. First, it removes the live range of the spilled virtual from
// the edge liveness results. Next, it removes the virtual from the list
// of cliques.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::clean_live_range(
    KillingSites_Fer & ks,
    unsigned sp_v,
    const MachineBasicBlock & mbb
) {
    // Kill the live range of v_sp, starting from the basic block where it is
    // defined.
    const MachineInstr * mi = this->def_use_sites->get_def_site(sp_v);
    const MachineBasicBlock * mbb_def = mi->getParent();
    // If the operand is a global variable, it may not have been defined. In
    // this case, nothing has to be done.
    if(mbb_def != NULL) {
        this->edge_liveness->kill_live_range(sp_v, * mbb_def);
    }
}


//===--------------------------------------------------------------------------
// This method converts the mapping from virtual register to physical register
// found by the perform_local_allocation method to a real assignment of
// physical registers in each operand of each instruction of the given basic
// block. Actually, it would be possible to implement this method directly in
// the perform_local_allocation function, but the separate implementation makes
// the program more readable.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::hard_code_register_mapping(MachineFunction & mf) {
    SSARegMap * srm = mf.getSSARegMap();
    unsigned v_reg = MRegisterInfo::FirstVirtualRegister;
    unsigned last_v_reg = srm->getLastVirtReg();
    while(v_reg <= last_v_reg) {
        unsigned p_reg = this->reg_mapping->get_physical_location(v_reg);
        if(p_reg != RegMapping_Fer::NO_REG) {
            this->vrm->assignVirt2Phys(v_reg, p_reg);
            v_reg++;
        }
    }
}


//===--------------------------------------------------------------------------
// This method calls the register allocation algorithm.
//===--------------------------------------------------------------------------
bool RegAllocChordal_Fer::runOnMachineFunction(MachineFunction & mf) {

    initialize_data_structures(mf);

    // Finds a mapping between virtual and physical registers:
    MachineBasicBlock *entry = mf.begin();
    std::set<MachineBasicBlock*> visited;
    for (df_ext_iterator<MachineBasicBlock*> dfi = df_ext_begin(entry, visited),
         end = df_ext_end(entry, visited); dfi != end; ++dfi) {
        MachineBasicBlock *mbb = *dfi;
        perform_local_allocation(*mbb);
    }

    // Destroy phi functions, by inserting copies and swaps whenever
    // appropriate:
    eliminate_phi_functions(mf);

    // Applies the mapping to the operands in the instructions:
    hard_code_register_mapping(mf);
    this->spiller->runOnMachineFunction(mf, *this->vrm);
    this->vrm.reset();

    delete_data_structures();

    return false;
}


//===--------------------------------------------------------------------------
// This function creates a register allocator based on the coloring of
// chordal graphs.
//===--------------------------------------------------------------------------
FunctionPass *llvm::createChordalRegisterAllocator() {
    return new RegAllocChordal_Fer();
}
