//===------------- DefUseSites_Fer.cpp - Printing prototype ---------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it stores the definition and use sites of virtual registers.
//
// Date: May 28th, 2006
//
//===----------------------------------------------------------------------===//
//
// This class implements a mapping of virtual registers to instructions.
// For each virtual, this class gives its definition site, which is unique,
// due to the SSA property, and all the use sites. A definition site is the
// instruction where the register is defined, and a use site is an
// instruction where the register is used. Such information allows to
// compute a weighting function used to choose the best candidate for
// spilling. The weighting function is given by the following algorithm:
//
// Instruction def_i := definition_site(v);
//
// Unsigned w := 1 + 10 ** loop_depth[def_i];
// 
// foreach ( Instruction i in use_sites(v) )
// |
// |    w := w + 1 + 10 ** loop_depth[i];
// |
// end foreach
//
// v2w(v) := w;
//
//===----------------------------------------------------------------------===//

#include <map>
#include <vector>
#include <sstream>                                // ostringstream
#include <iostream>                               // cerr

#include "llvm/CodeGen/Passes.h"                  // EdgeLivenessFernandoPassID
#include "llvm/Instructions.h"                    // MachineOperand
#include "llvm/Analysis/LoopInfo.h"
#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/CodeGen/MachineInstr.h"
#include "llvm/CodeGen/MachineFunctionPass.h"     // MachineBasicBlock
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"

#include "llvm/fernando/NeatPrint_Fer.h"          // class definition
#include "llvm/fernando/DefUseSites_Fer.h"        // class definition


using namespace llvm;

static RegisterAnalysis<DefUseSites_Fer>
X("Virtual2DefUseMapping", "Virtual to def/use mapping - Fernando.");


//===--------------------------------------------------------------------------
// This method returns true if the given operand has a definition site marked
// in this class; it returns false otherwise.
//===--------------------------------------------------------------------------
const bool DefUseSites_Fer::has_def_site(const unsigned v_reg) const {
    assert(MRegisterInfo::isVirtualRegister(v_reg) &&
                                      "Seeking def site of physical operand.");
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    if(this->v2def.size() < base_reg) {
        if(this->v2def[base_reg] != NULL) {
            return true;
        }
    }
    return false;
}


//===--------------------------------------------------------------------------
// Returns the instruction where the temporary is defined. This instruction is
// unique, because the target program is expected to be in the SSA-form.
//===--------------------------------------------------------------------------
const MachineInstr * DefUseSites_Fer::get_def_site(const unsigned v_reg) {
    assert(MRegisterInfo::isVirtualRegister(v_reg) &&
                                      "Seeking def site of physical operand.");
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    assert(this->v2def.size() > base_reg &&
              "Operand has no definition site. Maybe it is a spilled variable");
    assert(this->v2def[base_reg] != NULL &&
               "Operand has no definition site. Maybe it is a global variable");
    return this->v2def[base_reg];
}


//===--------------------------------------------------------------------------
// Returns the number of the block where the machine operand was defined.
//===--------------------------------------------------------------------------
int DefUseSites_Fer::get_def_block(const unsigned v_reg) {
    assert(MRegisterInfo::isVirtualRegister(v_reg) &&
                                      "Seeking def site of physical operand.");
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    assert(this->v2def.size() > base_reg &&
              "Operand has no definition site. Maybe it is a spilled variable");
    assert(this->v2def[base_reg] != NULL &&
               "Operand has no definition site. Maybe it is a global variable");
    const MachineInstr * mi = this->v2def[base_reg];
    return mi->getParent()->getNumber();
}


//===--------------------------------------------------------------------------
// Returns a vector containing all the instruction that use the given virtual
// register.
//===--------------------------------------------------------------------------
std::vector< MachineInstr * > &
DefUseSites_Fer::get_use_sites(const unsigned v_reg) {

    assert(MRegisterInfo::isVirtualRegister(v_reg) &&
                                      "Seeking def site of physical operand.");
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    assert(this->v2def.size() > base_reg &&
              "Operand has no definition site. Maybe it is a spilled variable");
    assert(this->v2def[base_reg] != NULL &&
               "Operand has no definition site. Maybe it is a global variable");

    return this->v2uses[base_reg];
}


//===--------------------------------------------------------------------------
// Computes the e-th power of n, that is, n**e;
//===--------------------------------------------------------------------------
const unsigned DefUseSites_Fer::pow(unsigned n, unsigned e) const {
    unsigned p = 1;
    for(unsigned i = 0; i < e; i++) {
        p = p * n;
    }
    return p;
}


//===--------------------------------------------------------------------------
// Returns the spilling weight of the given temporary. Virtual registers with
// smaller weights are better candidates to be spilled. The heuristic function
// is computed according to the following algorithm:
//
// Instruction def_i := definition_site(v);
//
// Unsigned w := 1 + 10 ** loop_depth[def_i];
// 
// foreach ( Instruction i in use_sites(v) )
// |
// |    w := w + 1 + 10 ** loop_depth[i];
// |
// end foreach
//
// v2w(v) := w;
// 
//===--------------------------------------------------------------------------
const unsigned DefUseSites_Fer::get_weight(unsigned v_reg) {
    const MachineInstr * def_site = get_def_site(v_reg);
    const MachineBasicBlock * mbb_def = def_site->getParent();
    unsigned w = pow(10, this->loop_depth[mbb_def->getNumber()]);
    std::vector< MachineInstr * > use_sites = get_use_sites(v_reg);
    for(unsigned i = 0; i < use_sites.size(); i++) {
        MachineInstr * use_site = use_sites[i];
        MachineBasicBlock * mbb_use = use_site->getParent();
        w = w + pow(10, this->loop_depth[mbb_use->getNumber()]);
    }
    return w;
}


//===--------------------------------------------------------------------------
// Free the memory, so the operating system can use it in other applications.
// TODO: it is necessary to verify if I am freeing the memory correctly.
//===--------------------------------------------------------------------------
void DefUseSites_Fer::releaseMemory() {
    this->loop_depth.clear();
    this->v2def.clear();
    this->v2uses.clear();
}


//===--------------------------------------------------------------------------
// This method visits a basic block and collects all the uses and defs present
// in this basic block.
//===--------------------------------------------------------------------------
void DefUseSites_Fer::visit_basic_block(MachineBasicBlock & mbb) {

    MachineBasicBlock::iterator mii = mbb.begin();
    while (mii != mbb.end()) {
        MachineInstr *mi = mii++;
        for (unsigned i = 0; i != mi->getNumOperands(); ++i) {
            MachineOperand & mo = mi->getOperand(i);
            if (mo.isRegister() && mo.getReg()) {
                if(MRegisterInfo::isVirtualRegister(mo.getReg())) {
                    unsigned v_reg = mo.getReg();
                    unsigned base_reg =
                                   v_reg - MRegisterInfo::FirstVirtualRegister;
                    if (mo.isDef()) {
                        this->v2def[base_reg] = mi;
                    } else {
                        this->v2uses[base_reg].push_back(mi);
                    }
                }
            }
        }
    }
}


//===--------------------------------------------------------------------------
// Prints in the standard error device all the information contained in this
// data structure.
//===--------------------------------------------------------------------------
void DefUseSites_Fer::dump() const {
    for(unsigned u = 0; u < this->v2def.size(); u++) {
        if(this->v2def[u]) {
            const MachineInstr * defi = this->v2def[u];
            const MachineBasicBlock * mbb_def = defi->getParent();
            unsigned reg_w = pow(10, this->loop_depth[mbb_def->getNumber()]);
            std::cerr << "- - - - - - - - - - - - - - - - - - - - - - - - -\n";
            std::cerr << "Def   : " << 
            NeatPrint_Fer::opcode2string(*defi, *mbb_def->getParent()) << "\n";
            std::cerr << ", Block " << mbb_def->getNumber() << "\n";
            std::vector< MachineInstr * > u_sites = this->v2uses[u];
            for(unsigned i = 0; i < u_sites.size(); i++) {
                MachineInstr * u_site = u_sites[i];
                MachineBasicBlock * mbb_u = u_site->getParent();
                std::cerr << "Use   : " <<
                   NeatPrint_Fer::opcode2string(* u_site, * mbb_u->getParent())
                                                                        <<"\n";
                std::cerr << ", Block " << mbb_u->getNumber() << "\n";
                reg_w = reg_w + pow(10, this->loop_depth[mbb_u->getNumber()]);
            }
            std::cerr << "Weight: " << reg_w << "\n";
        }
    }
}


//===--------------------------------------------------------------------------
// Adds to the loop_depth vector the nested factor of the basic block mbb. The
// nested factor is computed by the LoopInfo analysis, which is available in
// the li variable.
//===--------------------------------------------------------------------------
void DefUseSites_Fer::set_loop_depth(MachineBasicBlock & mbb, LoopInfo & li) {
    const BasicBlock * bb = mbb.getBasicBlock();
    unsigned nested_factor = li.getLoopDepth(bb);
    unsigned mbb_number = mbb.getNumber();

    if(this->loop_depth.size() <= mbb_number) {
        this->loop_depth.resize(mbb_number + 1);
    }

    this->loop_depth[mbb_number] = nested_factor;
}


//===--------------------------------------------------------------------------
// runOnMachineFunction - collects the use/def sites of each register, and
// computes their weights.
//===--------------------------------------------------------------------------
bool DefUseSites_Fer::runOnMachineFunction(MachineFunction & mf) {

    LoopInfo * loop_info = & getAnalysis<LoopInfo>();
    this->target_machine = & mf.getTarget();
    SSARegMap * srm = mf.getSSARegMap();
    unsigned last_v_reg = srm->getLastVirtReg() + 1;
    unsigned num_v_regs = last_v_reg - MRegisterInfo::FirstVirtualRegister;

    this->v2def.resize(num_v_regs + 1);
    this->v2uses.resize(num_v_regs + 1);

    for (MachineFunction::iterator mbb = mf.begin(), mbbe = mf.end();
                                                          mbb != mbbe; ++mbb) {
        visit_basic_block(*mbb);
        set_loop_depth(*mbb, *loop_info);
    }

    return false;
}
