//===------------- PhiCollector_Fer.cpp - Printing prototype --------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it groups virtual registers that are related by phi functions.
//
// Date: August 1th, 2006
//
//===----------------------------------------------------------------------===//
//
// This class groups together virtual registers that are related by phi
// functions in the control flow graph. Such virtuals describe the same
// variable in the original program. A good register allocator should try to
// assign the same physical register to these virtuals.
//
//===----------------------------------------------------------------------===//

#include <vector>
#include <sstream>                               // ostringstream
#include <iostream>

#include "llvm/CodeGen/Passes.h"                 // EdgeLivenessFernandoPassID
#include "llvm/CodeGen/MachineInstr.h"
#include "llvm/CodeGen/MachineFunctionPass.h"     // MachineBasicBlock
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"

#include "llvm/fernando/PhiCollector_Fer.h"      // class definition


using namespace llvm;

static RegisterAnalysis<PhiCollector_Fer>
X("PhiCollector", "Groups phi-related virtuals - Fernando.", true);


//===--------------------------------------------------------------------------
// This method returns true if the given virtual (or, for the same effect,
// any other phi-related virtual) has been mapped to a stack slot.
//===--------------------------------------------------------------------------
bool PhiCollector_Fer::has_stack_slot(unsigned v) {
    unsigned base_reg = v - MRegisterInfo::FirstVirtualRegister;
    if(base_reg >= this->v2s.size()) {
        return false;
    } else
        return this->v2s[base_reg] != -1;
}


//===--------------------------------------------------------------------------
// This method returns the stack slot that has been assigned to virtual
// v, if it has been spilled to memory. All the phi-related virtuals have
// the same stack slot.
//===--------------------------------------------------------------------------
int PhiCollector_Fer::get_stack_slot(unsigned v) {
    unsigned base_reg = v - MRegisterInfo::FirstVirtualRegister;
    if(base_reg >= this->v2s.size()) {
        return -1;
    } else
        return this->v2s[base_reg];
}


//===--------------------------------------------------------------------------
// This method maps the virtual v to the stack slot s. It also has the
// effect of assigning s to all other virtuals that are phi-related to v.
// If v has already been mapped to a stack slot, the function will produce
// an error.
//===--------------------------------------------------------------------------
void PhiCollector_Fer::map_to_stack_slot(unsigned v, int s) {
    unsigned base_reg = v - MRegisterInfo::FirstVirtualRegister;
    assert(base_reg < this->v2s.size() &&
                      "Index of register out of bounds of stack slot vector.");
    assert(s >= 0 && "Invalid stack slot.");
    assert(this->v2s[base_reg] == -1 &&
                     "Error: virtual has already been assigned a stack slot.");
    this->v2s[base_reg] = s;
    std::vector<bool> rel = this->related[base_reg];
    for(unsigned u = 0; u < rel.size(); u++) {
        if(rel[u] == true) {
            assert((this->v2s[u] == -1 || this->v2s[u] == s) &&
             "Error: related virtual has been assigned different stack slot.");
            this->v2s[u] = s;
        }
    }
}


//===--------------------------------------------------------------------------
// runOnMachineFunction - Groups together the phi related virtuals.
//===--------------------------------------------------------------------------
bool PhiCollector_Fer::runOnMachineFunction(MachineFunction & mf) {
    for (MachineFunction::const_iterator mbbi = mf.begin(), mbbe = mf.end();
                                                        mbbi != mbbe; ++mbbi) {
        const MachineBasicBlock * mbb = mbbi;
        MachineBasicBlock::const_iterator mii = mbb->begin();
        while (mii != mbb->end()) {
            const MachineInstr * mi = mii;
            if(mi->getOpcode() == TargetInstrInfo::PHI) {
                add_as_related( * mi );
            } else {
                // Always safe, because phi functions are always the first
                // instructions in the basic block.
                break;
            }
            mii++;
        }
    }
    return false;
}


//===--------------------------------------------------------------------------
// This method reads an instruction, and adds all the virtuals in this
// instruction to the same related set. If the instruction is not a phi
// function, an error is returned.
//===--------------------------------------------------------------------------
void PhiCollector_Fer::add_as_related(const MachineInstr & phi) {
    assert(phi.getOpcode() == TargetInstrInfo::PHI &&
                             "Only phi instructions define related virtuals.");
    std::vector<unsigned> vec;
    unsigned max_reg = 0;

    for(unsigned i = 0; i != phi.getNumOperands(); i++) {
        const MachineOperand & mo = phi.getOperand(i);
        if (mo.isRegister() && mo.getReg()) {
            unsigned v_reg = mo.getReg() - MRegisterInfo::FirstVirtualRegister;
            vec.push_back(v_reg);
            if(v_reg > max_reg) {
                max_reg = v_reg;
            }
        }
    }

    // Change the size of the related vector to include the new registers.
    if(this->related.size() <= max_reg) {
        this->related.resize(max_reg + 1);
        this->v2s.resize(max_reg + 1, -1);
    }

    // The vec array now contains all the variables in the phi function.
    // It is possible that some of them are related to other variables, not
    // present in the phi function. Therefore, the list of related variables
    // must be increased.
    unsigned old_size = vec.size();
    for(unsigned t = 0; t < old_size; t++) {
        // vec[t] hasn't even been added to related, so, continue
        if(this->related.size() <= vec[t]) {
            continue;
        } else {
            if(this->related[vec[t]].size() > 0) {
                for(unsigned u = 0; u < related[vec[t]].size(); u++) {
                    if(related[vec[t]][u]) {
                        vec.push_back(u);
                        if(u > max_reg) {
                            max_reg = u;
                        }
                    }
                }
            }
        }
    }

    // Add all the virtuals in the vec array to the same related set:
    for(unsigned t = 0; t < vec.size(); t++) {
        if(this->related[vec[t]].size() < max_reg) {
            this->related[vec[t]].resize(max_reg + 1, false);
        }
        for(unsigned u = 0; u < vec.size(); u++) {
            this->related[vec[t]][vec[u]] = true;
        }
    }
}


//===--------------------------------------------------------------------------
// This method returns true if v_reg is part of a phi function somewhere
// in the control flow graph, and false otherwise. 
//===--------------------------------------------------------------------------
bool PhiCollector_Fer::is_phi_related(unsigned v_reg) {
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    if(base_reg >= this->related.size()) {
        return false;
    } else
        return this->related[base_reg].size() > 0;
}


//===--------------------------------------------------------------------------
// This method adds "new_reg" to the set of related values of "v_reg".
//===--------------------------------------------------------------------------
void PhiCollector_Fer::add_to_related_set(unsigned v_reg, unsigned new_reg) {
    unsigned base_v = v_reg - MRegisterInfo::FirstVirtualRegister;
    unsigned base_n = new_reg - MRegisterInfo::FirstVirtualRegister;
    assert(this->related[base_v].size() > 0 && "No phi related variable");
    if(this->related.size() <= base_n) {
        this->related.resize(base_n + 1);
        this->v2s.resize(base_n + 1, -1);
    }
    this->related[base_n].resize(this->related.size(), false);
    unsigned old_size = this->related[base_v].size();
    for(unsigned u = 0; u < old_size; u++) {
        if(this->related[base_v][u]) {
            if(this->related[u].size() <= base_n) {
                this->related[u].resize(base_n + 1, false);
            }
            this->related[u][base_n] = true;
            this->related[base_n][u] = true;
        }
    }
}


//===--------------------------------------------------------------------------
// Prints the set of phi-related virtuals.
//===--------------------------------------------------------------------------
void PhiCollector_Fer::dump() {
    std::vector<bool> aux(this->related.size(), false);
    for(unsigned u = 0; u < this->related.size(); u++) {
        if(this->related[u].size() > 0 && aux[u] == false) {
            aux[u] = true;
            std::cerr << "[" << this->v2s[u] << "]  {";
            for(unsigned t = 0; t < this->related[u].size(); t++) {
                if(this->related[u][t]) {
                    std::cerr << " t" << t;
                    aux[t] = true;
                }
            }
            std::cerr << "}\n";
        }
    }
}
