//===------------ PhiDeconstruction.cpp - Phi-elimination Pass ------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it eliminate phi functions after register allocation has been
// performed.
//
// Date: June 30th, 2006
//
//===----------------------------------------------------------------------===//
//
// Traditional algorithms used to destroy phi functions normally replace them
// by copy instructions. Although the classical algorithms preserve the
// semantics of the SSA-form program, they may increase the number of registers
// found during the register assignment phase. To avoid this problem, we
// replace phi functions by parallel copies. Parallel copies are normally
// implemented with xor instructions (Remember the trick of swaping two
// variables without an auxiliary using xor?). The register swap is a costly
// operation, because it demands three instructions to be implemented;
// therefore, it should be avoided whenever possible. One way of avoiding
// swaps is by means of move instructions. Basically, swap operations are not
// necessary in every time that phi instructions are destroyed. This happens
// when the definition vector contains machine registers not used in the
// parameter vector. This pass tries to use such optimization  before inserting
//  swap operations. It performs phi elimination in time linear in the size of
// the definition vector.
//
//===----------------------------------------------------------------------===//


#include <vector>
#include <iostream>                              // cerr

#include "VirtRegMap.h"
#include "llvm/Instructions.h"                   // MachineOperand, etc
#include "llvm/CodeGen/MachineFunctionPass.h"    // MachineBasicBlock
#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/Target/TargetInstrInfo.h"

#include "llvm/fernando/PhiEq_Fer.h"
#include "llvm/fernando/EdgeLiveness_Fernando.h"
#include "llvm/fernando/PhiDeconstruction_Fer.h"
#include "llvm/fernando/RegMapping_Fer.h"

using namespace llvm;


//===--------------------------------------------------------------------------
// This method inserts a move instruction at the end of basic block mbb.
// The move instruction copies the value of register src to register dst.
// If the source of the move is a spill, then, instead of inserting a move
// instruction, this method directly loads the spilled value into the destiny.
//===--------------------------------------------------------------------------
void PhiDeconstruction_Fer::add_move(
    MachineBasicBlock & mbb,
    unsigned src,
    unsigned dst
) {
    assert(mbb.succ_size() == 1 && "Attempt to insert move in critical edge.");

    MachineBasicBlock::iterator iter = mbb.getFirstTerminator();

    const MRegisterInfo * reg_info =
                         this->machine_function->getTarget().getRegisterInfo();

    const TargetRegisterClass *rc =
                      this->machine_function->getSSARegMap()->getRegClass(src);

    if(src > this->reg_mapping->get_last_non_spill()) {
        int slot = this->vrm->getStackSlot(src);
        reg_info->loadRegFromStackSlot(mbb, iter, dst, slot, rc);
    } else {
        reg_info->copyRegToReg(mbb, iter, dst, src, rc);
    }

}


//===--------------------------------------------------------------------------
// This method inserts a sequence of instructions at the end of basic
// block mbb in order to swap the values of registers r1 and r2. Which
// instructions are inserted dependes on the target architecture.
// Normally swaps are implemented with three xor instructions.
//===--------------------------------------------------------------------------
void PhiDeconstruction_Fer::add_swap
                          (MachineBasicBlock & mbb, unsigned r1, unsigned r2) {

    assert(mbb.succ_size() == 1 && "Attempt to insert swap in critical edge.");

    const TargetMachine & tm = this->machine_function->getTarget();
    const TargetInstrInfo * tii = tm.getInstrInfo();

    /// Finds the first non branch instruction in the basic block.
    MachineBasicBlock::iterator iter = mbb.getFirstTerminator();

    const TargetRegisterClass *rc =
                       this->machine_function->getSSARegMap()->getRegClass(r1);

    const MRegisterInfo * reg_info =
                         this->machine_function->getTarget().getRegisterInfo();
    reg_info->swapRegs(mbb, iter, r1, r2, rc);
}


//===--------------------------------------------------------------------------
// This method fills vector r with indices that contain valid registers in v1
// and do not contain valid registers in v2. That is, this method fills vector
// r with the registers that are not used in v2, and that are used in v1. It
// implements set intersection in linear time.
//===--------------------------------------------------------------------------
void PhiDeconstruction_Fer::colors_v1_minus_colors_v2 (
    UnsignedVector_Fer & v1,
    UnsignedVector_Fer & v2,
    std::vector<ValueIndex> & r
) {
    r.clear();
    for(unsigned u = 0; u < v1.size(); u++) {
        unsigned value = v1.get(u);
        if(v1.get(u) == RegMapping_Fer::NO_REG
                                      || v1.get(u) == RegMapping_Fer::SP_REG) {
            continue;
        }
        bool is_present = false;
        for(unsigned t = 0; t < v2.size(); t++) {
            if(v2.get(t) == value) {
                is_present = true;
                break;
            }
        }
        if( ! is_present ) {
            unsigned index = u;
            ValueIndex vi(value, index);
            r.push_back(vi);
        }
    }
}


//===--------------------------------------------------------------------------
// This method inserts a store instruction at the end of the basic block mbb.
// This instruction stores the virtual use_v to the stack position assigned to
// the virtual def_v.
//===--------------------------------------------------------------------------
void PhiDeconstruction_Fer::insert_store(
    MachineBasicBlock & mbb,
    unsigned use_v,
    unsigned def_v
) {
    // Gets the stack slot where the virtual is mapped.
    assert(this->vrm->hasStackSlot(def_v) &&
                "Attempt to store a virtual that does not have a stack slot.");
    int slot = this->vrm->getStackSlot(def_v);

    const MRegisterInfo * ri =
                         this->machine_function->getTarget().getRegisterInfo();

    // gets the register class. Each virtual is part of a class that defines
    // the physical registers that can be assigned to it:
    const TargetRegisterClass *rc =
                    this->machine_function->getSSARegMap()->getRegClass(use_v);

    // Finds the first non branch instruction in the bb.
    MachineBasicBlock::iterator it = mbb.getFirstTerminator();

    ri->storeRegToStackSlot(mbb, it, use_v, slot, rc);
}


//===--------------------------------------------------------------------------
// The method assumes that use_v and def_v have been mapped to memory. It
// inserts instructions to tranfer the contents of use_v's location to def_v's
// location.
//===--------------------------------------------------------------------------
bool PhiDeconstruction_Fer::insert_memory_transfer(
    MachineBasicBlock & mbb,
    unsigned src,
    unsigned dst
) {
    unsigned p_reg = get_free_register(src);
    if(p_reg == RegMapping_Fer::NO_REG) {
        return false;
    }
    // Map the src and dst register to p_reg
    this->reg_mapping->set_color_spilled_register(src, p_reg);
    this->reg_mapping->set_color_spilled_register(dst, p_reg);

    // Gets the stack slots where the src and dst registers are mapped.
    assert(this->vrm->hasStackSlot(src) &&
                 "Attempt to load a virtual that does not have a stack slot.");
    assert(this->vrm->hasStackSlot(dst) &&
                "Attempt to store a virtual that does not have a stack slot.");

    const MRegisterInfo * ri =
                         this->machine_function->getTarget().getRegisterInfo();

    // gets the register class. Each virtual is part of a class that defines
    // the physical registers that can be assigned to it:
    const TargetRegisterClass * rc_src =
                      this->machine_function->getSSARegMap()->getRegClass(src);

    // Finds the first non branch instruction in the bb.
    MachineBasicBlock::iterator it = mbb.getFirstTerminator();

    ri->copyRegToReg(mbb, it, dst, src, rc_src);

    return true;
}


//===--------------------------------------------------------------------------
// This method inserts instructions to transfer a value in memory to another
// memory location.
//===--------------------------------------------------------------------------
bool PhiDeconstruction_Fer::allocate_memory_transfers(
    UnsignedVector_Fer & done_def,
    UnsignedVector_Fer & done_use,
    MachineBasicBlock & origin,
    PhiEq_Fer & phi_eq
) {
    bool result = true;
    for(unsigned idx_def = 0; idx_def < done_def.size(); idx_def++) {
        if(done_def.get(idx_def) == RegMapping_Fer::SP_REG &&
                             done_use.get(idx_def) == RegMapping_Fer::SP_REG) {
            unsigned vdef = phi_eq.get_def_virt(idx_def);
            unsigned vuse = phi_eq.get_use_virt(idx_def, origin);
            result = result && insert_memory_transfer(origin, vuse, vdef);
            if(result) {
                done_def.remove_index(idx_def);
                done_use.remove_index(idx_def);
            }
        }
    }
    return result;
}


//===--------------------------------------------------------------------------
// This method store virtuals in the use set of the phi function when the
// corresponding definition is a spilled value. This method is always safe,
// and can be called any time.
//===--------------------------------------------------------------------------
void PhiDeconstruction_Fer::resolve_virt2mem(
    UnsignedVector_Fer & done_def,
    UnsignedVector_Fer & done_use,
    MachineBasicBlock & origin,
    PhiEq_Fer & phi_eq
) {
    for(unsigned u = 0; u < done_def.size(); u++) {
        if(done_def.get(u) == RegMapping_Fer::SP_REG &&
                                   done_use.get(u) != RegMapping_Fer::SP_REG) {
            unsigned def_v = phi_eq.get_def_virt(u);
            unsigned use_v = phi_eq.get_use_virt(u, origin);
            insert_store(origin, use_v, def_v);
            done_def.remove_index(u);
            done_use.remove_index(u);
        }
    }
}


//===--------------------------------------------------------------------------
// This method helps destroying phi functions by using move instructions
// whenever applicable. Move instructions are preferable instead of swaps
// because generally they can be implemented cheaply. If def vector contains
// virtual v, which is not in the use vector, it is always safe to add a move
// into v. The basic block mbb is the block where the move instruction should
// be inserted.
//===--------------------------------------------------------------------------
void PhiDeconstruction_Fer::insert_move_instructions(
    MachineBasicBlock & mbb,
    std::vector<ValueIndex> & dif,
    UnsignedVector_Fer & done_def,
    UnsignedVector_Fer & done_use,
    PhiEq_Fer & phi_eq
) {
    std::vector<unsigned> phi_phys_prm = phi_eq.get_use_p_vec(mbb);
    colors_v1_minus_colors_v2(done_def, done_use, dif);
    while(dif.size() > 0) {
        ValueIndex def = dif.back();
        dif.pop_back();
        unsigned reg_def = def.value;
        unsigned idx_def = def.index;
        unsigned reg_use = phi_phys_prm[idx_def];
        unsigned virt_def = phi_eq.get_def_virt(idx_def);
        unsigned virt_use = phi_eq.get_use_virt(idx_def, mbb);
        add_move(mbb, virt_use, virt_def);
        done_def.remove_index(idx_def);
        done_use.remove_index(idx_def);
        if(done_def.contains(reg_use) && ! done_use.contains(reg_use)) {
            ValueIndex vi(reg_use, done_def.get_index(reg_use));
            dif.push_back(vi);
        }
    }
}


//===--------------------------------------------------------------------------
// This method inserts swap instructions in order to destroy phi-functions.
// A swap instruction is implemented with xor instructions, and each swap
// will produce three xor's. Therefore, move instructions should be inserted
// whenever possible; however, in (a, b) := (b, a), move instructions cannot
// be used without demanding an extra register.
//===--------------------------------------------------------------------------
void PhiDeconstruction_Fer::insert_swap_instructions(
    MachineBasicBlock & mbb,
    std::vector<ValueIndex> & dif,
    UnsignedVector_Fer & done_def,
    UnsignedVector_Fer & done_use,
    PhiEq_Fer & phi_eq
) {
    std::vector<unsigned> def_v;
    std::vector<unsigned> phi_phys_prm = phi_eq.get_use_p_vec(mbb);
    done_def.fill_vector_indices(this->num_p_regs, def_v);
    std::vector<unsigned> use_v;
    done_use.fill_vector_indices(this->num_p_regs, use_v);
    while(done_def.has_virtual()) {
        unsigned reg_def = done_def.choose_and_remove_any_virtual();
        unsigned idx_def = def_v[reg_def];
        unsigned reg_use = phi_phys_prm[idx_def];
        unsigned idx_use = use_v[reg_def];
        if(reg_def != reg_use) {
            unsigned virt_def = phi_eq.get_use_virt(idx_def, mbb);
            unsigned virt_use = phi_eq.get_use_virt(idx_use, mbb);
            assert(this->reg_mapping->compatible_class(virt_use, reg_def) &&
                  "Swaping incompatible registers during phi deconstruction.");
            add_swap(mbb, virt_def, virt_use); // add swap inst.
            use_v[reg_def] = idx_def;
            use_v[reg_use] = idx_use;
            phi_phys_prm[idx_def] = reg_def;
            phi_phys_prm[idx_use] = reg_use;
            phi_eq.swap_virtuals(idx_def, idx_use, mbb);
        }
    }
}


//===--------------------------------------------------------------------------
// This method determines if p_reg is not currently allocated to any virtual
// register alive in the beginning of the given basic block.
//===--------------------------------------------------------------------------
bool PhiDeconstruction_Fer::is_register_available(unsigned p_reg) {
    if (this->occupied_physicals[p_reg] == true) {
        return false;
    }
    const TargetMachine & target_machine = this->machine_function->getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    for (const unsigned *a = reg_info->getAliasSet(p_reg); *a; ++a) {
        if (this->occupied_physicals[*a] == true) {
            return false;
        }
    }
    return true;
}


//===--------------------------------------------------------------------------
// This method returns a register that can be assigned to the given
// virtual register. It will return RegMapping_Fer::NO_REG if there
// is no register available.
//===--------------------------------------------------------------------------
unsigned PhiDeconstruction_Fer::get_free_register(unsigned v_reg) {
    const TargetRegisterClass *trc =
                    this->machine_function->getSSARegMap()->getRegClass(v_reg);
    TargetRegisterClass::iterator trci =
                         trc->allocation_order_begin(* this->machine_function);
    TargetRegisterClass::iterator trce =
                           trc->allocation_order_end(* this->machine_function);
    for (; trci != trce; ++trci) {
        if (is_register_available(*trci)) {
            return *trci;
        }
    }
    return RegMapping_Fer::NO_REG;
}


//===--------------------------------------------------------------------------
// This method determines which virtual registers are reaching this basic block
// alive. Let L be the set of live registers: L = live_edge(org, dst) -
// live_use + live_def.
//===--------------------------------------------------------------------------
void PhiDeconstruction_Fer::set_occupied_regs(
    MachineBasicBlock & org,
    MachineBasicBlock & dst,
    UnsignedVector_Fer & def,
    UnsignedVector_Fer & use
) {
    this->occupied_physicals.assign(this->num_p_regs, false);
    // Compute live_edge(org, dst)
    const std::vector<bool> & live_v =
                             this->edge_liveness->get_live_variables(org, dst);
    for(unsigned u = 0; u < live_v.size(); u++) {
        if(live_v[u]) {
            unsigned v_reg = u + MRegisterInfo::FirstVirtualRegister;
            unsigned p_reg = this->reg_mapping->get_physical_location(v_reg);
            assert(p_reg < this->occupied_physicals.size() &&
                                    "Virtual has no valid physical location.");
            this->occupied_physicals[p_reg] = true;
        }
    }
    // Remove live_use:
    for(unsigned u = 0; u < use.size(); u++) {
        unsigned p_reg = use.get(u);
        if(p_reg < this->occupied_physicals.size()) {
            this->occupied_physicals[p_reg] = false;
        } else {
            assert(p_reg == RegMapping_Fer::SP_REG ||
                        p_reg == RegMapping_Fer::NO_REG && "Invalid register");
        }
    }
    // Add live_def:
    for(unsigned u = 0; u < def.size(); u++) {
        unsigned p_reg = def.get(u);
        if(p_reg < this->occupied_physicals.size()) {
            this->occupied_physicals[p_reg] = false;
        } else {
            assert(p_reg == RegMapping_Fer::SP_REG ||
                        p_reg == RegMapping_Fer::NO_REG && "Invalid register");
        }
    }
}


//===--------------------------------------------------------------------------
// This method determines which virtual registers are reaching this basic block
// alive. These virtuals are alive in the edge that goes from src basic block
// to dst basic block.
//===--------------------------------------------------------------------------
void PhiDeconstruction_Fer::set_occupied_regs(
    MachineBasicBlock & org,
    MachineBasicBlock & dst
) {
    this->occupied_physicals.assign(this->num_p_regs, false);

    const std::vector<bool> & live_v =
                             this->edge_liveness->get_live_variables(org, dst);
    for(unsigned u = 0; u < live_v.size(); u++) {
        if(live_v[u]) {
            unsigned v_reg = u + MRegisterInfo::FirstVirtualRegister;
            unsigned p_reg = this->reg_mapping->get_physical_location(v_reg);
            assert(p_reg < this->occupied_physicals.size() &&
                                    "Virtual has no valid physical location.");
            this->occupied_physicals[p_reg] = true;
        }
    }
}


//===--------------------------------------------------------------------------
// This method inserts move and swap instructions in the control flow graph,
// in order to reproduce the semantics of phi-functions. Notice that the
// phi functions are not eliminated by this method.
//===--------------------------------------------------------------------------
void PhiDeconstruction_Fer::insert_moves_and_swaps(
    MachineBasicBlock & mbb
) {
    PhiEq_Fer phi_eq(mbb, * this->vrm, * this->reg_mapping);

    if(phi_eq.number_phi_functions() == 0) {
        return;
    }

    const std::vector<unsigned> phi_def = phi_eq.get_def_phys();

    // Insert new instructions in each of the predecessor blocks
    for(MachineBasicBlock::pred_iterator pred_aux = mbb.pred_begin();
                                      pred_aux != mbb.pred_end(); pred_aux++) {
        MachineBasicBlock* pred_mbb = *pred_aux;
        std::vector<unsigned> phi_phys_prm = phi_eq.get_use_p_vec(*pred_mbb);

        // These vector avoid writing on the original vector of the phi
        // equation, so that other iterations of the loop don't get a messed
        // phi definition vector.
        UnsignedVector_Fer done_def(phi_def);
        UnsignedVector_Fer done_use(phi_phys_prm);

        // First of all, try to find free registers for the memory transfers
        // In this initial stage, the busy registers are the register alive
        // across the phi equation, and the registers alive in phi_phys_prm.
        set_occupied_regs(* pred_mbb, mbb);
        allocate_memory_transfers(done_def, done_use, *pred_mbb, phi_eq);

        // Now, reset the occupied physical registers. Occupied = live_edge -
        // done_use + done_def;
        set_occupied_regs(* pred_mbb, mbb, done_def, done_use);

        // If there are [spill] := non_spill, it is always safe to store
        // the non_spill in the previous basic block:
        resolve_virt2mem(done_def, done_use, *pred_mbb, phi_eq);

        // This vector will hold the virtuals whose physical registers are
        // present in done_def, but not in done_use
        std::vector<ValueIndex> dif;

        // Before using swaps, attempt to insert move instructions every time
        // it is possible:
        insert_move_instructions(*pred_mbb, dif, done_def, done_use, phi_eq);

        // Third, insert swaps:
        insert_swap_instructions(*pred_mbb, dif, done_def, done_use, phi_eq);

        // After inserting moves and swaps, the only thing that may remain are
        // pairs of virtuals mapped to memory.
        bool success =
              allocate_memory_transfers(done_def, done_use, *pred_mbb, phi_eq);

        // This is a temporary solution. Even if there are no free registers
        // at all, it is still possible to allocate the memory transfers. But,
        // in this case, a occupied register would have to be spilled.
        // TODO: fix this for an unsuccessful case.
        assert(success &&
                        "Unable to find a free register for memory transfer.");
    }
}


//===--------------------------------------------------------------------------
// This method removes all the occurrences of phi-functions from the
// given basic block. Before this method is called, instructions must be
// inserted in the control flow graph to reproduce the behavior of phi
// functions.
//===--------------------------------------------------------------------------
void PhiDeconstruction_Fer::remove_phi_functions(MachineBasicBlock & mbb) {

    if(!mbb.empty()) {
        while (mbb.front().getOpcode() == TargetInstrInfo::PHI) {
            MachineInstr * mphi = mbb.remove(mbb.begin());
            delete mphi;
        }
    }

/*
    // The infamous code that has costed me hours of pain: I am adding it here
    // for pedagogic purposes only.
    MachineBasicBlock::iterator mii = mbb.begin();
    while (mii != mbb.end() && mii->getOpcode() == TargetInstrInfo::PHI) {
        MachineInstr * mi = mii;
        mi->eraseFromParent();
        mii++;
    }
*/
}


//===--------------------------------------------------------------------------
// runOnMachineFunction - Invokes the phi elimination algorithm. Normally this
// method would be private, and would be called by the pass manager, but in the
// current implementation, I don't really now how to implement this class as a
// pass.
//===--------------------------------------------------------------------------
bool PhiDeconstruction_Fer::runOnMachineFunction(
    MachineFunction & mf
) {
    //std::cerr << "Phi elimination phase.\n";

    this->machine_function = & mf;

    // Get the number of machine registers:
    const TargetMachine & target_machine = mf.getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    this->num_p_regs = reg_info->getNumRegs();
    this->occupied_physicals.resize(this->num_p_regs);

    for (MachineFunction::iterator mbbi = mf.begin(), mbbe = mf.end();
                                                        mbbi != mbbe; ++mbbi) {
        MachineBasicBlock & mbb = *mbbi;
        insert_moves_and_swaps(mbb);
        remove_phi_functions(mbb);
    }

    this->occupied_physicals.clear();

    return false;
}
