//===-------------------- PhiEq_Fer.cpp - PHI Matrix ----------------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This class represents the phi functions in the header of a basic block.
//
// Date: Jun 30th, 2006
//
//===----------------------------------------------------------------------===//
//
// This class implements a phi equation. A phi matrix equation such as
// A <=phi= M is an abstraction used to represent the phi functions that are
// present in the beginning of a basic block. The vector in the left side
// denotes the variables defined by the phi functions. The matrix in the right
// represents the parameters of the phi functions. For each predecessor basic
// block, the matrix M contains a column, that hold the variables alive if the
// flow reaches the current block coming from that predecessor. Ex.:
//
// a       | a_1 a_2 a_3 |
// b =phi= | b_1 b_2 b_3 |
// c       | c_1 c_2 c_3 |
//
// This matrix represents the three phi functions: a = (a1,a2,a3),
// b = (b1,b2,b3) and c = (c1,c2,c3).
//
//===----------------------------------------------------------------------===//

#include <map>
#include <sstream>

#include "llvm/Instructions.h"                   // MachineOperand, etc
#include "llvm/CodeGen/MachineFunctionPass.h"    // MachineBasicBlock
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "VirtRegMap.h"

#include "llvm/fernando/RegMapping_Fer.h"
#include "llvm/fernando/NeatPrint_Fer.h"
#include "llvm/fernando/PhiEq_Fer.h"

using namespace llvm;

//===--------------------------------------------------------------------------
// Constructor method. Creates a new matrix equation and fills it up with
// the phi functions defined in the given basic block. This method
// does not fills the virtual mapping in the phi equation.
//===--------------------------------------------------------------------------
PhiEq_Fer::PhiEq_Fer(const MachineBasicBlock & mbb) {
    for(MachineBasicBlock::const_iterator bbi=mbb.begin(), bbe = mbb.end();
          bbi != bbe && bbi->getOpcode() == TargetInstrInfo::PHI; ++bbi ) {
        const MachineInstr * phi = bbi;
        unsigned def_reg = phi->getOperand(0).getReg();
        this->def_phys.push_back(def_reg);
        for ( int i = bbi->getNumOperands() - 1; i >= 2; i -= 2 ) {
            unsigned use_reg = phi->getOperand(i-1).getReg();
            MachineBasicBlock & pred =
                                *phi->getOperand(i).getMachineBasicBlock();
            unsigned mbb_num = pred.getNumber();
            if(this->prm_phys.find(mbb_num) == this->prm_phys.end()) {
                std::vector<unsigned> col;
                this->prm_phys[mbb_num] = col;
                this->col2mbb.push_back(mbb_num);
            }
            std::vector<unsigned> & col = this->prm_phys[mbb_num];
            col.push_back(use_reg);
        }
    }
}


//===--------------------------------------------------------------------------
// Second constructor method. Creates a new matrix equation and fills it
// up with the phi functions defined in the given basic block. The
// difference for the first constructor is that, in this case, the values
// used to fill the phi function are given by the register mapping
// object.
//===--------------------------------------------------------------------------
PhiEq_Fer::PhiEq_Fer(
    const MachineBasicBlock & mbb,
    const VirtRegMap & vrm,
    RegMapping_Fer & rm
) {
    for(MachineBasicBlock::const_iterator bbi=mbb.begin(), bbe = mbb.end();
          bbi != bbe && bbi->getOpcode() == TargetInstrInfo::PHI; ++bbi ) {
        const MachineInstr * phi = bbi;
        unsigned op_def = phi->getOperand(0).getReg();
        unsigned def_reg;
        if(vrm.hasStackSlot(op_def)) {
            def_reg = RegMapping_Fer::SP_REG;
        } else {
            def_reg = rm.get_physical_location(op_def);
        }
        this->def_phys.push_back(def_reg);
        this->def_virt.push_back(op_def);
        for ( int i = bbi->getNumOperands() - 1; i >= 2; i -= 2 ) {
            unsigned use_virt = phi->getOperand(i-1).getReg();
            unsigned use_phys;
            // Even when the virtual register is a spill, it is possible
            // that it is assigned a physical register; however, the phi
            // elimination algorithm needs that the virtual's location be
            // different of any valid physical register.
            if(vrm.hasStackSlot(use_virt)) {
                use_phys = RegMapping_Fer::SP_REG;
            } else {
                use_phys = rm.get_physical_location(use_virt);
            }
            MachineBasicBlock & pred =
                                *phi->getOperand(i).getMachineBasicBlock();
            unsigned mbb_num = pred.getNumber();
            if(this->prm_phys.find(mbb_num) == this->prm_phys.end()) {
                // create vector for physical registers
                std::vector<unsigned> col_phys;
                this->prm_phys[mbb_num] = col_phys;
                // create vector for virtual registers
                std::vector<unsigned> col_virt;
                this->prm_virt[mbb_num] = col_virt;
                // store the index of this machine basic block
                this->col2mbb.push_back(mbb_num);
            }
            std::vector<unsigned> & col_phys = this->prm_phys[mbb_num];
            col_phys.push_back(use_phys);
            std::vector<unsigned> & col_virt = this->prm_virt[mbb_num];
            col_virt.push_back(use_virt);
        }
    }
}


//===--------------------------------------------------------------------------
// Third constructor method. It creates a new phi equation from a second
// one. The valid vector marks which phi functions in ph must be added to
// the new equation.
//===--------------------------------------------------------------------------
PhiEq_Fer::PhiEq_Fer(PhiEq_Fer & ph, std::vector<bool> valid) {
    assert(valid.size() == ph.number_phi_functions() &&
                                "Invalid vector in PhiEq_Fer constructor");
    for(unsigned u = 0; u < ph.def_phys.size(); u++) {
        if(valid[u]) {
            this->def_phys.push_back(ph.def_phys[u]);
            this->def_virt.push_back(ph.def_virt[u]);
        }
    }
   for(unsigned u = 0; u < ph.col2mbb.size(); u++) {
        this->col2mbb.push_back(ph.col2mbb[u]);
    }
    std::map< unsigned, std::vector<unsigned> >::iterator it;
    for(it = ph.prm_phys.begin(); it != ph.prm_phys.end(); it++) {
        std::vector<unsigned> vec1 = (*it).second;
        std::vector<unsigned> vec2;
        assert(vec1.size() == valid.size() &&
                            "Invalid use vector in PhiEq_Fer constructor");
        for(unsigned v = 0; v < valid.size(); v++) {
            if(valid[v]) {
                vec2.push_back(vec1[v]);
            }
        }
        this->prm_phys[(*it).first] = vec2;
    }
    for(it = ph.prm_virt.begin(); it != ph.prm_virt.end(); it++) {
        std::vector<unsigned> vec1 = (*it).second;
        std::vector<unsigned> vec2;
        assert(vec1.size() == valid.size() &&
                            "Invalid use vector in PhiEq_Fer constructor");
        for(unsigned v = 0; v < valid.size(); v++) {
            if(valid[v]) {
                vec2.push_back(vec1[v]);
            }
        }
        this->prm_virt[(*it).first] = vec2;
    }
}


//===--------------------------------------------------------------------------
// Each column of the phi matrix represents an execution path. This method
// returns the set of variables defined in the execution path from the
// pred block and the current block.
//===--------------------------------------------------------------------------
std::vector<unsigned>& PhiEq_Fer::get_use_p_vec(const MachineBasicBlock &pred){
    unsigned mbb_num = pred.getNumber();
    return this->prm_phys[mbb_num];
}


//===--------------------------------------------------------------------------
// This method swaps the two virtual registers in positions i1 and i2 of
// the vector that reaches this phi equation flowing from block pred.
//===--------------------------------------------------------------------------
void PhiEq_Fer::swap_virtuals(
    unsigned i1,
    unsigned i2,
    const MachineBasicBlock & pred
) {
    unsigned mbb_num = pred.getNumber();
    std::vector<unsigned> & vec = this->prm_virt[mbb_num];
    unsigned aux = 0;
    aux = vec[i1];
    vec[i1] = vec[i2];
    vec[i2] = aux;
}


//===--------------------------------------------------------------------------
// This method returns the left side of the matrix equation, that is, the
// vector of variables defined by the set of phi functions that this
// equation represents.
//===--------------------------------------------------------------------------
const std::vector<unsigned> & PhiEq_Fer::get_def_phys() {
    return this->def_phys;
}


//===--------------------------------------------------------------------------
// This method returns the number of phi functions present in this
// matrix equation. Each phi function defines a different variable.
//===--------------------------------------------------------------------------
unsigned PhiEq_Fer::number_phi_functions() const {
    return this->def_phys.size();
}


//===--------------------------------------------------------------------------
// This method returns the number of predecessor blocks that gave origin
// to this matrix equation. Each predecessor block contains the
// live ranges of the variables represented in one of the columns of the
// matrix equation.
//===--------------------------------------------------------------------------
unsigned PhiEq_Fer::ncols() const {
    return this->col2mbb.size();
}


//===--------------------------------------------------------------------------
// Returns a string describing this phi matrix:
// For this set of phi functions: a = (a1,a2,a3), b = (b1,b2,b3) and
// c = (c1,c2,c3), the following matrix will be printed:
//
// a <= [ a_1 a_2 a_3 ]
// b <= [ b_1 b_2 b_3 ]
// c <= [ c_1 c_2 c_3 ]
//===--------------------------------------------------------------------------
std::string PhiEq_Fer::to_string() {
    std::ostringstream aux;
    for(unsigned u = 0; u < this->def_phys.size(); u++) {
        aux << this->def_phys[u] << " <= [";
        for(unsigned t = 0; t < this->col2mbb.size(); t++) {
            unsigned mbb_num = this->col2mbb[t];
            std::vector<unsigned> & col = this->prm_phys[mbb_num];
            aux << " " << col[u];
        }
        aux << " ]\n";
    }
    return aux.str();
}


//===--------------------------------------------------------------------------
// This method prints a formated string. The string will be printed on a
// box of 'packSize'. The extra spaces will be filled with c.
//===--------------------------------------------------------------------------
std::string PhiEq_Fer::print_pck_str(std::string str, int packSize, char c) {
    std::ostringstream aux;
    int strSize = str.size();
    int incSize = strSize < packSize ? packSize - strSize : 0 ;
    std::string strAux(incSize, c);
    aux << str << strAux;
    return aux.str() ;
}


//===--------------------------------------------------------------------------
// This method returns a string describing this phi matrix. The difference
// for the parameterless to_string method is that in this case the name
// of the register is printed, instead of its code.
//===--------------------------------------------------------------------------
std::string PhiEq_Fer::to_string(MachineFunction & mf) {
    std::ostringstream aux;
    const TargetMachine & target_machine = mf.getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();

    for(unsigned u = 0; u < this->def_phys.size(); u++) {
        if(def_phys[u] == RegMapping_Fer::SP_REG) {
            aux << print_pck_str("X", 8, ' ') << " <= [";
        } else {
            aux << print_pck_str
                (NeatPrint_Fer::op2string(def_phys[u], mf), 8, ' ') << " <= [";
        }
        for(unsigned t = 0; t < this->col2mbb.size(); t++) {
            unsigned mbb_num = this->col2mbb[t];
            std::vector<unsigned> & col = this->prm_phys[mbb_num];
            if(col[u] == RegMapping_Fer::SP_REG) {
                aux << " " << print_pck_str("X", 8, ' ');
            } else {
                aux << " " <<
                 print_pck_str(NeatPrint_Fer::op2string(col[u], mf),8,' ');
            }
        }
        aux << " ]\n";
    }
    return aux.str();
}


//===--------------------------------------------------------------------------
// This method returns the virtual register that is mapped to the given
// index in the definition vector of the phi equation.
//===--------------------------------------------------------------------------
unsigned PhiEq_Fer::get_def_virt(unsigned index) const {
    assert(index < this->def_virt.size() && "Invalid index phi_eq.");
    return this->def_virt[index];
}


//===--------------------------------------------------------------------------
// This method returns the virtual register that is mapped to the given
// index in the use vector that corresponds to the given basic block.
//===--------------------------------------------------------------------------
unsigned PhiEq_Fer::get_use_virt(unsigned index, const MachineBasicBlock &mbb){
    unsigned mbb_num = mbb.getNumber();
    assert(index < this->prm_virt[mbb_num].size() &&
                                 "Invalid index in use vector of phi_eq.");
    return this->prm_virt[mbb_num][index];
}
