//===--------------- RegClique_Fer.cpp - Set of Registers -----------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This class implements a set of registers. It distinguises between virtual
// and machine registes.
//
// Date: May 27th, 2006
//
//===----------------------------------------------------------------------===//
//
// This class implements a set of registers. It is mostly used to represent
// cliques of registers. A clique of register is a set of register whose live
// ranges overlap at the same point in the control flow graph of the SSA-form
// program.
//===----------------------------------------------------------------------===//

#include "llvm/Instructions.h"                   // MachineOperand, etc
#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/CodeGen/MachineFunctionPass.h"    // MachineBasicBlock
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"

#include "llvm/fernando/EdgeLiveness_Fernando.h"
#include "llvm/fernando/NeatPrint_Fer.h"
#include "llvm/fernando/RegClique_Fer.h"

using namespace llvm;


//===--------------------------------------------------------------------------
// Constructor method. It will initialize this register set with the out
// set of the given machine block. The out set of a basic block is the set of
// variables alive at the end of the basic block. It can be obtained by the
// union of live sets of all the edges that leave the basic block. Just a
// remark, the parameters are not changed in the scope of this function, but
// I could not declared them cons becaue some methods invoked on them
// discard the const qualifier.
//===--------------------------------------------------------------------------
RegClique_Fer::RegClique_Fer(
                              MachineFunction & mf,
                              const MachineBasicBlock & mbb,
                              const EdgeLiveness_Fernando * edge_liveness
) {
    this->phys_size = 0;
    this->virt_size = 0;
    this->target_machine = & mf.getTarget();
    // discover the number of virtual registers:
    SSARegMap * srm = mf.getSSARegMap();
    unsigned last_v_reg = srm->getLastVirtReg() + 1;
    unsigned nv = last_v_reg - MRegisterInfo::FirstVirtualRegister;
    this->virt_reg.assign(nv, false);
    this->m_reg_info = this->target_machine->getRegisterInfo();
    this->phys_reg.assign(this->m_reg_info->getNumRegs(), false);
    this->allocatable_regs = this->m_reg_info->getAllocatableSet(mf);

    for(MachineBasicBlock::const_succ_iterator succ_aux = mbb.succ_begin();
                                      succ_aux != mbb.succ_end(); succ_aux++) {
        const MachineBasicBlock* succ_mbb = *succ_aux;
        const std::vector<bool> & succ_live =
                        edge_liveness->get_live_variables(mbb, *succ_mbb);
        for (unsigned i = 0; i < succ_live.size(); i++) {
            if(succ_live[i]) {
                if(this->virt_reg[i] == false) {
                    this->virt_size++;
                }
                this->virt_reg[i] = true;
            }
        }
    }
    // Finally, if the last block in the function is a return, make sure to
    // add the live out set of the function to this set of registers.
    const TargetInstrInfo & tii = * (this->target_machine->getInstrInfo());
    if (!mbb.empty() && tii.isReturn(mbb.const_back().getOpcode())) {
        for (MachineFunction::liveout_iterator iter = mf.liveout_begin(),
                                 end = mf.liveout_end(); iter != end; ++iter) {
            assert(MRegisterInfo::isPhysicalRegister(*iter) &&
                                   "Cannot have a live-out virtual register!");
            if(this->phys_reg[*iter] == false) {
                this->phys_size++;
            }
            this->phys_reg[*iter] = true;
        }
    }
}


//===--------------------------------------------------------------------------
// Constructor method. It will initialize this register set with the virtual
// registers alive in the edge that links mbb1 to mbb2. Notice that if mbb2
// is not a successor of mbb1, no register will be in this register set.
//===--------------------------------------------------------------------------
RegClique_Fer::RegClique_Fer(
                              MachineFunction & mf,
                              const MachineBasicBlock & mbb1,
                              const MachineBasicBlock & mbb2,
                              const EdgeLiveness_Fernando * edge_liveness
) {
    this->phys_size = 0;
    this->virt_size = 0;
    this->target_machine = & mf.getTarget();
    // discover the number of virtual registers:
    SSARegMap * srm = mf.getSSARegMap();
    unsigned last_v_reg = srm->getLastVirtReg() + 1;
    unsigned nv = last_v_reg - MRegisterInfo::FirstVirtualRegister;
    this->virt_reg.assign(nv, false);
    this->m_reg_info = this->target_machine->getRegisterInfo();
    this->phys_reg.assign(this->m_reg_info->getNumRegs(), false);
    this->allocatable_regs = this->m_reg_info->getAllocatableSet(mf);

    const std::vector<bool> & live_set =
                                 edge_liveness->get_live_variables(mbb1, mbb2);
    for (unsigned i = 0; i < live_set.size(); i++) {
        if(live_set[i]) {
            if(this->virt_reg[i] == false) {
                this->virt_size++;
            }
            this->virt_reg[i] = true;
        }
    }
}


//===--------------------------------------------------------------------------
// Constructor method. It produces a set of registers that is a copy of the set
// initially given as parameter.
//===--------------------------------------------------------------------------
RegClique_Fer::RegClique_Fer(RegClique_Fer & rc) {
    this->phys_size = rc.phys_size;
    this->virt_size = rc.virt_size;
    this->virt_reg.assign(rc.virt_reg.begin(), rc.virt_reg.end());
    this->phys_reg.assign(rc.phys_reg.begin(), rc.phys_reg.end());
    this->allocatable_regs.assign
                      (rc.allocatable_regs.begin(), rc.allocatable_regs.end());
    this->target_machine = rc.target_machine;
    this->m_reg_info = rc.m_reg_info;
}


//===--------------------------------------------------------------------------
// Returns a new set of registers that is an intersection between this set and
// the given set.
//===--------------------------------------------------------------------------
RegClique_Fer * RegClique_Fer::intersection(RegClique_Fer & rc) const {
    RegClique_Fer * aux = new RegClique_Fer(rc);
    // Resolve the physical registers first:
    for(unsigned u = 0; u < this->phys_reg.size(); u++) {
        if(this->phys_reg[u] == false) {
            aux->phys_reg[u] = false;
        }
    }
    // Now, resolve the virtual registers:
    for(unsigned u = 0; u < this->virt_reg.size(); u++) {
        if(this->virt_reg[u] == false) {
            aux->virt_reg[u] = false;
        }
    }
    return aux;
}


//===--------------------------------------------------------------------------
// This method removes the register given as parameter from this register set.
// It will return true if the register was present in the set, and false
// otherwise.
//===--------------------------------------------------------------------------
bool RegClique_Fer::remove_reg(unsigned reg_num) {
    if(MRegisterInfo::isVirtualRegister(reg_num)) {
        unsigned base_reg = reg_num - MRegisterInfo::FirstVirtualRegister;
        assert(this->virt_reg.size() > base_reg &&
               "RegClique_Fer: not valid temporary register.");
        if(this->virt_reg[base_reg]) {
            this->virt_reg[base_reg] = false;
            this->virt_size--;
            return true;
        }
    } else if(MRegisterInfo::isPhysicalRegister(reg_num)) {
        if(this->phys_reg.size() > reg_num) {
            if(this->phys_reg[reg_num]) {
                this->phys_reg[reg_num] = false;
                this->phys_size--;
                return true;
            }
        }
    }
    return false;
}


//===--------------------------------------------------------------------------
// This method returns a reference to the vector that represents the virtual
// registers stored in this set. This vector cannot be modified outside this
// class.
//===--------------------------------------------------------------------------
const std::vector<bool> & RegClique_Fer::get_virtual_vector() const {
    return this->virt_reg;
}


//===--------------------------------------------------------------------------
// Return an array containing all the registers in ascending order. The first
// elements will represent physical registers, because, in LLVM, physical
// registers are denoted by integer numbers from 0 to K, where K is the number
// of physical registers.
//===--------------------------------------------------------------------------
void RegClique_Fer::to_vector(std::vector<unsigned> & regs) const {
    regs.clear();
    for(unsigned i = 0; i < this->phys_reg.size(); i++) {
        if(this->phys_reg[i]) {
            regs.push_back(i);
        }
    }
    for(unsigned i = 0; i < this->virt_reg.size(); i++) {
        if(this->virt_reg[i]) {
            regs.push_back(i + MRegisterInfo::FirstVirtualRegister);
        }
    }
}


//===--------------------------------------------------------------------------
// This method fills the given vector with all the physical registers present
// in this set.
//===--------------------------------------------------------------------------
void RegClique_Fer::p2vector(std::vector<unsigned> & regs) const {
    regs.resize(this->phys_size);
    for(unsigned i = 0, j = 0; i < this->phys_reg.size(); i++) {
        if(this->phys_reg[i]) {
            regs[j] = i;
            j++;
        }
    }
}


//===--------------------------------------------------------------------------
// This method fills the given vector with all the virtual registers
// present in the set.
//===--------------------------------------------------------------------------
void RegClique_Fer::v2vector(std::vector<unsigned> & regs) const {
    regs.resize(this->virt_size);
    for(unsigned i = 0, j = 0; i < this->virt_reg.size(); i++) {
        if(this->virt_reg[i]) {
            regs[j] = i + MRegisterInfo::FirstVirtualRegister;
            j++;
        }
    }
}


//===--------------------------------------------------------------------------
// This method returns true if the register set contains the register given as
// parameter. It will return false otherwise. The caller must make sure that
// the parameter describes a valid register number.
//===--------------------------------------------------------------------------
bool RegClique_Fer::contains_reg(unsigned reg_num) const {

    if(MRegisterInfo::isVirtualRegister(reg_num)) {
        unsigned base_reg = reg_num - MRegisterInfo::FirstVirtualRegister;
        assert(this->virt_reg.size() > base_reg &&
               "RegClique_Fer: not valid temporary register.");
        return this->virt_reg[base_reg];
    } else if(MRegisterInfo::isPhysicalRegister(reg_num)) {
        if(this->allocatable_regs[reg_num]) {
            assert(this->phys_reg.size() > reg_num &&
                               "RegClique_Fer: not valid temporary register.");
            return this->phys_reg[reg_num];
        }
    }
    return false;
}


//===--------------------------------------------------------------------------
// This method adds the register given as parameter to this register set.
//===--------------------------------------------------------------------------
void RegClique_Fer::add_reg(unsigned reg_num) {
    if(MRegisterInfo::isVirtualRegister(reg_num)) {
        unsigned base_reg = reg_num - MRegisterInfo::FirstVirtualRegister;
        assert(this->virt_reg.size() > base_reg &&
               "RegClique_Fer: not valid temporary register.");
        if(this->virt_reg[base_reg] == false) {
            this->virt_size++;
        }
        this->virt_reg[base_reg] = true;
    } else if(MRegisterInfo::isPhysicalRegister(reg_num)) {
        if(this->allocatable_regs[reg_num]) {
            if (this->phys_reg.size() <= reg_num) {
                this->phys_reg.resize(reg_num + 1);
            }
            if(this->phys_reg[reg_num] == false) {
                this->phys_size++;
            }
            this->phys_reg[reg_num] = true;
        }
    }
}


//===--------------------------------------------------------------------------
// This method grows the size of the set of virtual registers. Notice that it
// only GROWS the size, it never decrease it. Therefore, if the new size is
// smaller than the old one, the size will not change.
//===--------------------------------------------------------------------------
void RegClique_Fer::grow_virtual(unsigned new_size) {
    while(this->virt_reg.size() <= new_size) {
        this->virt_reg.push_back(false);
    }
}


//===--------------------------------------------------------------------------
// This method adds each register uses in the instruction to this set of
// registers. Both, explicit and implicit uses are added.
//===--------------------------------------------------------------------------
void RegClique_Fer::add_uses(const MachineInstr & mi) {
    // handle explicit uses
    bool isPhi = (mi.getOpcode() == TargetInstrInfo::PHI) ? true : false ;
    for (unsigned i = 0; i != mi.getNumOperands(); i++) {
        const MachineOperand & mo = mi.getOperand(i);
        if (mo.isUse() && mo.isRegister() && mo.getReg() && (!isPhi) ) {
            add_reg(mo.getReg());
        }
    }
    // handle implicit uses
    const TargetInstrInfo & tii = * (this->target_machine->getInstrInfo());
    const TargetInstrDescriptor &tid = tii.get(mi.getOpcode());
    for (const unsigned *implicit_use = tid.ImplicitUses;
                                               *implicit_use; ++implicit_use) {
        add_reg(*implicit_use);
    }
}


//===--------------------------------------------------------------------------
// This method removes each register defined in the instruction to this set of
// registers. Both, explicit and implicit definitions are removed.
//===--------------------------------------------------------------------------
void RegClique_Fer::remove_definitions(const MachineInstr & mi) {
    // handle explicit defs
    for (unsigned i = 0; i != mi.getNumOperands(); i++) {
        const MachineOperand & mo = mi.getOperand(i);
        if (mo.isDef() && mo.isRegister() && mo.getReg()) {
            remove_reg(mo.getReg());
        }
    }
    // handle implicit defs
    const TargetInstrInfo & tii = * (this->target_machine->getInstrInfo());
    const TargetInstrDescriptor &tid = tii.get(mi.getOpcode());
    for (const unsigned *implicit_def = tid.ImplicitDefs;
                                               *implicit_def; ++implicit_def) {
        remove_reg(*implicit_def);
    }
}


//===--------------------------------------------------------------------------
// Returns a string in the format (r1, r2, ..., rm)+(t1, t2, ..., tn), where
// ri represents a machine register, and t1 represents a temporary (also called
// virtual) register.
//===--------------------------------------------------------------------------
std::string RegClique_Fer::to_string() const {
    bool isFirst = true;
    std::ostringstream aux;

    // first, generate a string for the physical registers:
    aux << "(";
    for (unsigned i = 0, e = this->phys_reg.size(); i != e; ++i) {
        if(this->phys_reg[i]) {
            if(isFirst) {
                isFirst = false;
                aux << "r" << i;
            } else {
                aux << ", " << "r" << i;
            }
        }
    }

    // Now, take care of the virtual registers:
    isFirst = true;
    aux << ")+(";
    for (unsigned i = 0, e = this->virt_reg.size(); i != e; ++i) {
        if(this->virt_reg[i]) {
            if(isFirst) {
                isFirst = false;
                aux << "t" << i;
            } else {
                aux << ", " << "t" << i;
            }
        }
    }

    aux << ")";
    return aux.str();
}
